#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
r"""
iCalendar parsing and generating.

iCalendar (RFC 2445) is a big and hard-to-read specification.  This module
supports only a subset of it: VEVENT and VFREEBUSY components with a limited
set of attributes and a limited recurrence model.  The subset should be
sufficient for interoperation with desktop calendaring applications like
Apple's iCal, Mozilla Calendar, Evolution and KOrganizer as well as the
Outlook meeting planner.

If you have a calendar, you can convert it to an iCalendar file like this:

    >>> from pytz import timezone
    >>> from datetime import datetime, timedelta
    >>> from schooltool.calendar.simple import ImmutableCalendar
    >>> from schooltool.calendar.simple import SimpleCalendarEvent
    >>> event = SimpleCalendarEvent(datetime(2004, 12, 16, 10, 58, 47),
    ...                             timedelta(hours=1), "doctests",
    ...                             location=u"Matar\u00f3",
    ...                             description=u"Writing doctests",
    ...                             unique_id="12345678-5432@example.com")
    >>> calendar = ImmutableCalendar([event])

    >>> ical_file_as_string = "\r\n".join(
    ...         convert_calendar_to_ical(calendar) + [''])

The returned string is in UTF-8.

    >>> event.location.encode("UTF-8") in ical_file_as_string
    True

You can also parse iCalendar files back into calendars:

    >>> event_iterator = read_icalendar(ical_file_as_string)
    >>> new_calendar = ImmutableCalendar(event_iterator)
    >>> [e.title for e in new_calendar]
    [u'doctests']
    >>> e == event
    True

There is some trickery to make empty calendars work:

    >>> empty_calendar = ImmutableCalendar([])
    >>> ical_file_as_string = "\r\n".join(
    ...         convert_calendar_to_ical(empty_calendar) + [''])

    >>> list(read_icalendar(ical_file_as_string))
    []

"""

import pytz
import datetime
import re
from cStringIO import StringIO

from schooltool.calendar.simple import SimpleCalendarEvent


EMPTY_CALENDAR_PLACEHOLDER = 'empty-calendar-placeholder@schooltool.org'

def convert_event_to_vfb(event):
    r"""Convert an ICalendarEvent to iCalendar FREEBUSY entry.

    Returns a list of strings (without newlines) in UTF-8.

        >>> from datetime import datetime, timedelta
        >>> event = SimpleCalendarEvent(datetime(2004, 12, 16, 10, 7, 29),
        ...                             timedelta(hours=1), "iCal rendering",
        ...                             location="Big room",
        ...                             description="Blah blah\nblah!",
        ...                             unique_id="12345678-5432@example.com")
        >>> lines = convert_event_to_vfb(event)
        >>> print "\n".join(lines)
        FREEBUSY:20041216T100729/20041216T11072900Z

    """
    # XXX: not handling recurrence yet
    # XXX: shouldn't assume event.dtstart is in UTC
    assert event.dtstart.tzname() == 'UTC'
    return ["FREEBUSY:%s/%s00Z" % (ical_datetime(event.dtstart),
                               ical_datetime(event.dtstart + event.duration))]


def convert_event_to_ical(event):
    r"""Convert an ICalendarEvent to iCalendar VEVENT component.

    Returns a list of strings (without newlines) in UTF-8.

        >>> from datetime import datetime, timedelta
        >>> event = SimpleCalendarEvent(datetime(2004, 12, 16, 10, 7, 29),
        ...                             timedelta(hours=1), "iCal rendering",
        ...                             location="Big room",
        ...                             description="Blah blah\nblah!",
        ...                             unique_id="12345678-5432@example.com")
        >>> lines = convert_event_to_ical(event)
        >>> print "\n".join(lines)
        BEGIN:VEVENT
        UID:12345678-5432@example.com
        SUMMARY:iCal rendering
        LOCATION:Big room
        DESCRIPTION:Blah blah\nblah!
        DTSTART:20041216T100729Z
        DURATION:PT1H
        DTSTAMP:...
        END:VEVENT

        >>> from schooltool.calendar.recurrent import DailyRecurrenceRule
        >>> event = SimpleCalendarEvent(datetime(2005, 2, 11, 22, 42, 50),
        ...                             timedelta(minutes=15), "iCal tests",
        ...                             recurrence=DailyRecurrenceRule(),
        ...                             unique_id="12345678-9876@example.com")
        >>> lines = convert_event_to_ical(event)
        >>> print "\n".join(lines)
        BEGIN:VEVENT
        UID:12345678-9876@example.com
        SUMMARY:iCal tests
        RRULE:FREQ=DAILY;INTERVAL=1
        DTSTART:20050211T224250Z
        DURATION:PT15M
        DTSTAMP:...
        END:VEVENT

    All-day events have DTSTART as a date:

        >>> from schooltool.calendar.recurrent import WeeklyRecurrenceRule
        >>> event = SimpleCalendarEvent(datetime(2005, 2, 11, 0, 0, 0),
        ...                             timedelta(days=2), "iCal tests",
        ...                             recurrence=WeeklyRecurrenceRule(),
        ...                             unique_id="12345678-9876@example.com",
        ...                             allday=True)
        >>> lines = convert_event_to_ical(event)
        >>> print "\n".join(lines)
        BEGIN:VEVENT
        UID:12345678-9876@example.com
        SUMMARY:iCal tests
        RRULE:FREQ=WEEKLY;BYDAY=FR;INTERVAL=1
        DTSTART;VALUE=DATE:20050211
        DURATION:P2D
        DTSTAMP:...
        END:VEVENT

    """
    dtstamp = datetime.datetime.utcnow().strftime("%Y%m%dT%H%M%SZ")
    result = [
        "BEGIN:VEVENT",
        "UID:%s" % ical_text(event.unique_id),
        "SUMMARY:%s" % ical_text(event.title)]
    if event.location:
        result.append("LOCATION:%s" % ical_text(event.location))
    if event.description:
        result.append("DESCRIPTION:%s" % ical_text(event.description))
    if event.recurrence is not None:
        start = event.dtstart
        result.extend(event.recurrence.iCalRepresentation(start))

    if event.allday:
        dtstart = 'DTSTART;VALUE=DATE:%s' % ical_date(event.dtstart)
    else:
        # XXX: shouldn't assume event.dtstart is in UTC
        assert event.dtstart.tzname() == 'UTC'
        dtstart = 'DTSTART:%sZ' % ical_datetime(event.dtstart)
    result += [dtstart,
               "DURATION:%s" % ical_duration(event.duration),
               "DTSTAMP:%s" % dtstamp,
               "END:VEVENT"]
    return result


def convert_calendar_to_vfb(calendar):
    r"""Convert an ICalendar to a VFREEBUSY component
    Returns a list of strings (without newlines) in UTF-8.  They should be
    joined with '\r\n' to get a valid iCalendar file.

        >>> from schooltool.calendar.simple import ImmutableCalendar
        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> from datetime import datetime, timedelta
        >>> event1 = SimpleCalendarEvent(datetime(2004, 11, 16, 10, 7, 29),
        ...                             timedelta(hours=1), "iCal rendering",
        ...                             location="Big room",
        ...                             unique_id="12345678-5432@example.com")
        >>> event2 = SimpleCalendarEvent(datetime(2004, 12, 16, 10, 7, 29),
        ...                             timedelta(hours=2), "iCal rendering",
        ...                             location="Big room",
        ...                             unique_id="12345678-9876@example.com")
        >>> calendar = ImmutableCalendar([event1, event2])
        >>> lines = convert_calendar_to_vfb(calendar)
        >>> print "\n".join(lines)
        BEGIN:VCALENDAR
        VERSION:2.0
        PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        METHOD:PUBLISH
        BEGIN:VFREEBUSY
        FREEBUSY:20041116T100729/20041116T11072900Z
        FREEBUSY:20041216T100729/20041216T12072900Z
        END:VFREEBUSY
        END:VCALENDAR

    XXX VFREEBUSY should just expand every occurence

        >>> lines = convert_calendar_to_vfb(ImmutableCalendar())
        >>> print "\n".join(lines)
        BEGIN:VCALENDAR
        VERSION:2.0
        PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        METHOD:PUBLISH
        BEGIN:VFREEBUSY
        END:VFREEBUSY
        END:VCALENDAR


    """
    header = ["BEGIN:VCALENDAR",
              "VERSION:2.0",
              "PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN",
              "METHOD:PUBLISH",
              "BEGIN:VFREEBUSY"]
    events = []
    for event in calendar:
        events += convert_event_to_vfb(event)
    footer = ["END:VFREEBUSY",
              "END:VCALENDAR"]
    return header + events + footer


def convert_calendar_to_ical(calendar):
    r"""Convert an ICalendar to iCalendar VCALENDAR component.

    Returns a list of strings (without newlines) in UTF-8.  They should be
    joined with '\r\n' to get a valid iCalendar file.

        >>> from schooltool.calendar.simple import ImmutableCalendar
        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> from datetime import datetime, timedelta
        >>> event = SimpleCalendarEvent(datetime(2004, 12, 16, 10, 7, 29),
        ...                             timedelta(hours=1), "iCal rendering",
        ...                             location="Big room",
        ...                             unique_id="12345678-5432@example.com")
        >>> calendar = ImmutableCalendar([event])
        >>> lines = convert_calendar_to_ical(calendar)
        >>> print "\n".join(lines)
        BEGIN:VCALENDAR
        VERSION:2.0
        PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        BEGIN:VEVENT
        UID:12345678-5432@example.com
        SUMMARY:iCal rendering
        LOCATION:Big room
        DTSTART:20041216T100729Z
        DURATION:PT1H
        DTSTAMP:...
        END:VEVENT
        END:VCALENDAR

    Empty calendars are not allowed by RFC 2445, so we have to invent a dummy
    event:

        >>> lines = convert_calendar_to_ical(ImmutableCalendar())
        >>> print "\n".join(lines)
        BEGIN:VCALENDAR
        VERSION:2.0
        PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN
        BEGIN:VEVENT
        UID:...
        SUMMARY:Empty calendar
        DTSTART:19700101T000000Z
        DURATION:P0D
        DTSTAMP:...
        END:VEVENT
        END:VCALENDAR

    """
    header = ["BEGIN:VCALENDAR",
              "VERSION:2.0",
              "PRODID:-//SchoolTool.org/NONSGML SchoolTool//EN"]
    footer = ["END:VCALENDAR"]
    events = []
    for event in calendar:
        events += convert_event_to_ical(event)
    if not events:
        placeholder = SimpleCalendarEvent(datetime.datetime(1970, 1, 1),
                                          datetime.timedelta(0),
                                          "Empty calendar",
                                          unique_id=EMPTY_CALENDAR_PLACEHOLDER)
        events += convert_event_to_ical(placeholder)
    return header + events + footer


def ical_text(value):
    r"""Format value according to iCalendar TEXT escaping rules.

    Converts Unicode strings to UTF-8 as well.

        >>> ical_text('Foo')
        'Foo'
        >>> ical_text(u'Matar\u00f3')
        'Matar\xc3\xb3'
        >>> ical_text('\\')
        '\\\\'
        >>> ical_text(';')
        '\\;'
        >>> ical_text(',')
        '\\,'
        >>> ical_text('\n')
        '\\n'
    """
    return (value.encode('UTF-8')
                 .replace('\\', '\\\\')
                 .replace(';', '\\;')
                 .replace(',', '\\,')
                 .replace('\n', '\\n'))


def ical_date(value):
    """Format a date as an iCalendar DATE value.

        >>> ical_date(datetime.date(2004, 3, 4))
        '20040304'
    """
    return value.strftime("%Y%m%d")


def ical_datetime(value):
    """Format a datetime as an iCalendar DATETIME value.

        >>> ical_datetime(datetime.datetime(2004, 12, 16, 10, 45, 07))
        '20041216T104507'

    """
    return value.strftime('%Y%m%dT%H%M%S')


def ical_duration(value):
    """Format a timedelta as an iCalendar DURATION value.

        >>> from datetime import timedelta
        >>> ical_duration(timedelta(11))
        'P11D'
        >>> ical_duration(timedelta(-14))
        '-P14D'
        >>> ical_duration(timedelta(1, 7384))
        'P1DT2H3M4S'
        >>> ical_duration(timedelta(1, 7380))
        'P1DT2H3M'
        >>> ical_duration(timedelta(1, 7200))
        'P1DT2H'
        >>> ical_duration(timedelta(0, 7200))
        'PT2H'
        >>> ical_duration(timedelta(0, 7384))
        'PT2H3M4S'
        >>> ical_duration(timedelta(0, 184))
        'PT3M4S'
        >>> ical_duration(timedelta(0, 22))
        'PT22S'
        >>> ical_duration(timedelta(0, 3622))
        'PT1H0M22S'

    """
    sign = ""
    if value.days < 0:
        sign = "-"
    timepart = ""
    if value.seconds:
        timepart = "T"
        hours = value.seconds // 3600
        minutes = value.seconds % 3600 // 60
        seconds = value.seconds % 60
        if hours:
            timepart += "%dH" % hours
        if minutes or (hours and seconds):
            timepart += "%dM" % minutes
        if seconds:
            timepart += "%dS" % seconds
    if value.days == 0 and timepart:
        return "%sP%s" % (sign, timepart)
    else:
        return "%sP%dD%s" % (sign, abs(value.days), timepart)


def read_icalendar(icalendar_text, charset='UTF-8', fallback_tz=pytz.utc):
    """Read an iCalendar file and return calendar events.

    Returns an iterator over calendar events.

    `icalendar_text` can be a file object or a string.  It is assumed that
    the iCalendar file contains UTF-8 text.

    `fallback_tz` is used for timestamps when the timezone is not specified
    or not recognized.

    Unsupported features of the iCalendar file (e.g. VTODO components, complex
    recurrence rules, unknown properties) are silently ignored.
    """

    if not hasattr(icalendar_text, 'read'):
        # It is not a file-like object -- let's assume it is a string
        icalendar_text = StringIO(icalendar_text)

    rows = RowParser.parse(icalendar_text, charset)
    vccol = VCalendarCollection.parse(rows)

    for calendar in vccol.vcalendars:
        for vevent in calendar.events:
            if vevent.uid == EMPTY_CALENDAR_PLACEHOLDER:
                continue # Ignore empty calendar placeholder "event"

            dtstart = vevent.dtstart
            if not isinstance(dtstart, datetime.datetime):
                dtstart = datetime.datetime.combine(dtstart,
                                                    datetime.time(0))
            # XXX regression test for events with a date as dtend
            dtend = vevent.dtend
            if not isinstance(dtend, datetime.datetime):
                dtend = datetime.datetime.combine(dtend,
                                                  datetime.time(0))

            if dtstart.tzinfo is None:
                dtstart_tz = calendar.timezones.get(vevent.dtstart_tzid, fallback_tz)
                dtstart = dtstart_tz.localize(dtstart).astimezone(pytz.utc)
            if dtend.tzinfo is None:
                dtend_tz = calendar.timezones.get(vevent.dtend_tzid, fallback_tz)
                dtend = dtend_tz.localize(dtend).astimezone(pytz.utc)
            duration = dtend - dtstart

            yield SimpleCalendarEvent(dtstart, duration,
                                      vevent.summary or '',
                                      location=vevent.location,
                                      description=vevent.description,
                                      unique_id=vevent.uid,
                                      recurrence=vevent.rrule,
                                      allday=vevent.all_day_event)


#
# The rest of this module could use some review and refactoring
#


ical_weekdays = ['MO', 'TU', 'WE', 'TH', 'FR', 'SA', 'SU']


class RowParser(object):

    @staticmethod
    def _parseRow(record_str):
        """Parse a single content line.

        A content line consists of a property name (optionally followed by a
        number of parameters) and a value, separated by a colon.  Parameters
        (if present) are separated from the property name and from each other
        with semicolons.  Parameters are of the form name=value; value
        can be double-quoted.

        Returns a tuple (name, value, param_dict).  Case-insensitive values
        (i.e. property names, parameter names, unquoted parameter values) are
        uppercased.

        Raises ICalParseError on syntax errors.

        >>> RowParser._parseRow('foo:bar')
        ('FOO', 'bar', {})
        >>> RowParser._parseRow('foo;value=bar:BAZFOO')
        ('FOO', 'BAZFOO', {'VALUE': 'BAR'})

        """

        it = iter(record_str)
        getChar = it.next

        def err(msg):
            raise ICalParseError("%s in line:\n%s" % (msg, record_str))

        try:
            c = getChar()
            # name
            key = ''
            while c.isalnum() or c == '-':
                key += c
                c = getChar()
            if not key:
                err("Missing property name")
            key = key.upper()
            # optional parameters
            params = {}
            while c == ';':
                c = getChar()
                # param name
                param = ''
                while c.isalnum() or c == '-':
                    param += c
                    c = getChar()
                if not param:
                    err("Missing parameter name")
                param = param.upper()
                # =
                if c != '=':
                    err("Expected '='")
                # value (or rather a list of values)
                pvalues = []
                while True:
                    c = getChar()
                    if c == '"':
                        c = getChar()
                        pvalue = ''
                        while c >= ' ' and c not in ('\177', '"'):
                            pvalue += c
                            c = getChar()
                        # value is case-sensitive in this case
                        if c != '"':
                            err("Expected '\"'")
                        c = getChar()
                    else:
                        # unquoted value
                        pvalue = ''
                        while c >= ' ' and c not in ('\177', '"', ';', ':',
                                                     ','):
                            pvalue += c
                            c = getChar()
                        pvalue = pvalue.upper()
                    pvalues.append(pvalue)
                    if c != ',':
                        break
                if len(pvalues) > 1:
                    params[param] = pvalues
                else:
                    params[param] = pvalues[0]
            # colon and value
            if c != ':':
                err("Expected ':'")
            value = ''.join(it)
        except StopIteration:
            err("Syntax error")
        else:
            return (key, value, params)

    @staticmethod
    def parse(file, charset='UTF-8'):
        """A generator that returns one record at a time, as a tuple of
        (name, value, params).
        """
        record = []
        for line in file:
            if not line.strip():
                continue
            if line[0] in '\t ':
                line = line[1:]
            elif record:
                row = "".join(record).decode(charset)
                yield RowParser._parseRow(row)
                record = []
            if line.endswith('\r\n'):
                record.append(line[:-2])
            elif line.endswith('\n'):
                # strictly speaking this is a violation of RFC 2445
                record.append(line[:-1])
            else:
                # strictly speaking this is a violation of RFC 2445
                record.append(line)
        if record:
            row = "".join(record).decode(charset)
            yield RowParser._parseRow(row)


def parse_block(rows):
    """Read a block from BEGIN until END.

    Returns a dict with sub blocks and a dict of properties.

        >>> parse_block([])
        ([], {})

        >>> lines = ["BEGIN:VTIMEZONE",
        ...          "SOME:Property",
        ...          "END:VTIMEZONE"]
        >>> parse_block(RowParser.parse(lines))
        ([], {u'VTIMEZONE': [[(u'BEGIN', u'VTIMEZONE', {}),
                              (u'SOME', u'Property', {}),
                              (u'END', u'VTIMEZONE', {})]]})

        >>> lines = ["BEGIN:VTIMEZONE",
        ...          "SOME:Property",
        ...          "END:VTIMEZONE",
        ...          "PROP1: VAL1",
        ...          "PROP2:",
        ...          " VAL2",
        ...          "prop3;value=bar:val3"]
        >>> props, blocks = parse_block(RowParser.parse(lines))
        >>> props
        [(u'PROP1', u' VAL1', {}),
         (u'PROP2', u'VAL2', {}),
         (u'PROP3', u'val3', {u'VALUE': u'BAR'})]
        >>> blocks
        {u'VTIMEZONE': [[(u'BEGIN', u'VTIMEZONE', {}),
                         (u'SOME', u'Property', {}),
                         (u'END', u'VTIMEZONE', {})]]}

        >>> lines = ["BEGIN:VTIMEZONE",
        ...          "SOME:Property",
        ...          "BEGIN:STANDARD",
        ...          "OTHER:Property",
        ...          "END:STANDARD",
        ...          "END:VTIMEZONE",
        ...          "BEGIN:VTIMEZONE",
        ...          "SOME:Property",
        ...          "BEGIN:DAYLIGHT",
        ...          "ANOTHER:Property",
        ...          "END:DAYLIGHT",
        ...          "END:VTIMEZONE",
        ...          "PROP1: VAL1",
        ...          "PROP2:",
        ...          " VAL2",
        ...          "prop3;value=bar:val3"]
        >>> props, blocks = parse_block(RowParser.parse(lines))
        >>> props
        [(u'PROP1', u' VAL1', {}),
         (u'PROP2', u'VAL2', {}),
         (u'PROP3', u'val3', {u'VALUE': u'BAR'})]
        >>> blocks
        {u'VTIMEZONE': [[(u'BEGIN', u'VTIMEZONE', {}),
                         (u'SOME', u'Property', {}),
                         (u'BEGIN', u'STANDARD', {}),
                         (u'OTHER', u'Property', {}),
                         (u'END', u'STANDARD', {}),
                         (u'END', u'VTIMEZONE', {})],
                        [(u'BEGIN', u'VTIMEZONE', {}),
                         (u'SOME', u'Property', {}),
                         (u'BEGIN', u'DAYLIGHT', {}),
                         (u'ANOTHER', u'Property', {}),
                         (u'END', u'DAYLIGHT', {}),
                         (u'END', u'VTIMEZONE', {})]]}

        >>> lines = ["BEGIN:VCALENDAR",
        ...          "BEGIN:VEVENT",
        ...          "DTSTART;VALUE=DATE:20010203"]
        >>> props, blocks = parse_block(RowParser.parse(lines))
        Traceback (most recent call last):
        ...
        ICalParseError: Mismatched BEGIN/END

    """
    blocks = {}
    props = []
    parse_stack = []
    block = []
    for key, value, params in rows:
        if key == "BEGIN":
            parse_stack.append(value)
            # if this is the first time we encounter this block in
            # top level
            if (len(parse_stack) == 1 and
                value not in blocks.keys()):
                blocks[value] = []
        if parse_stack:
            block.append((key, value, params))
        else:
            props.append((key, value, params))
        if key == "END":
            if not parse_stack or parse_stack[-1] != value:
                raise ICalParseError("Mismatched BEGIN/END")
            if len(parse_stack) == 1:
                blocks[parse_stack[0]].append(block)
                block = []
            parse_stack.pop(-1)
    if parse_stack:
        raise ICalParseError("Mismatched BEGIN/END")
    return props, blocks


class VCalendarCollection(object):

    def __init__(self, vcalendars):
        self.vcalendars = vcalendars

    @staticmethod
    def parse(rows):
        """Parse a list of rows into a VCalendarCollection object."""
        props, blocks = parse_block(rows)

        if props:
            raise ICalParseError("Text outside VCALENDAR component")

        # Handle empty ics files
        if not blocks:
            vcalendars = []
        elif 'VCALENDAR' not in blocks.keys():
            raise ICalParseError('This is not iCalendar')
        elif blocks.keys() != ['VCALENDAR']:
            raise ICalParseError("Text outside VCALENDAR component")
        else:
            vcalendars = [VCalendar.parse(block)
                          for block in blocks['VCALENDAR']]
        return VCalendarCollection(vcalendars)


class VCalendar(object):

    def __init__(self, events, timezones):
        self.events = events
        self.timezones = timezones

    @staticmethod
    def parse(rows):
        """Parse a list of rows into a VCalendar object."""
        props, blocks = parse_block(rows[1:-1])

        timezones = {'UTC': pytz.utc}
        for block in blocks.get('VTIMEZONE', []):
            vtimezone = VTimezone.parse(block)
            tzinfo = vtimezone.getTzinfo()
            if tzinfo is None:
                continue # couldn't parse it :/
                # TODO: the user should be informed
            timezones[vtimezone.tzid.upper()] = tzinfo
        events = [VEvent.parse(block)
                  for block in blocks.get('VEVENT', [])]
        return VCalendar(events, timezones)


class VTimezone(object):

    def __init__(self, tzid, tznames, x_lic_location=None):
        tzid = tzid.encode('US-ASCII')
        self.tzid = tzid
        self.tznames = tznames
        self.x_lic_location = x_lic_location

    def getTzinfo(self):
        """Deduce the pytz timezone from the VTIMEZONE block."""
        if self.tzid in pytz.all_timezones:
            return pytz.timezone(self.tzid)
        elif self.x_lic_location in pytz.all_timezones:
            return pytz.timezone(self.x_lic_location)
        else:
            for tzname in reversed(self.tznames):
                if tzname in pytz.all_timezones:
                    return pytz.timezone(tzname)
        # We don't know
        return None

    @staticmethod
    def parse(rows):
        """Parse a list of rows into a VTimezone object."""
        rows = list(rows)
        props, blocks = parse_block(rows[1:-1])

        tzid = None
        x_lic_location = None
        for key, value, params in props:
            if key == 'TZID':
                tzid = value
            elif key == 'X-LIC-LOCATION':
                x_lic_location = value
        if not tzid:
            raise ICalParseError("Missing TZID in VTIMEZONE block")

        tznames = []
        for block in blocks.get('STANDARD', []):
            for key, value, params in block:
                if key == 'TZNAME':
                    tznames.append(value)
        if not tznames:
            raise ICalParseError("Missing STANDARD section in VTIMEZONE block")

        return VTimezone(tzid, tznames, x_lic_location=x_lic_location)


def parse_text(value):
    r"""Parse iCalendar TEXT value.

        >>> parse_text('Foo')
        'Foo'
        >>> parse_text('\\\\')
        '\\'
        >>> parse_text('\\;')
        ';'
        >>> parse_text('\\,')
        ','
        >>> parse_text('\\n')
        '\n'
        >>> parse_text('A string with\\; some\\\\ characters\\nin\\Nit')
        'A string with; some\\ characters\nin\nit'
        >>> parse_text('Unterminated \\')
        Traceback (most recent call last):
          ...
        IndexError: string index out of range

    """
    if '\\' not in value:
        return value
    out = []
    prev = 0
    while True:
        idx = value.find('\\', prev)
        if idx == -1:
            break
        out.append(value[prev:idx])
        if value[idx + 1] in 'nN':
            out.append('\n')
        else:
            out.append(value[idx + 1])
        prev = idx + 2
    out.append(value[prev:])
    return "".join(out)


def parse_date(value):
    """Parse iCalendar DATE value.  Returns a date instance.

        >>> parse_date('20030405')
        datetime.date(2003, 4, 5)
        >>> parse_date('20030405T060708')
        Traceback (most recent call last):
          ...
        ValueError: Invalid iCalendar date: '20030405T060708'
        >>> parse_date('')
        Traceback (most recent call last):
          ...
        ValueError: Invalid iCalendar date: ''
        >>> parse_date('yyyymmdd')
        Traceback (most recent call last):
          ...
        ValueError: Invalid iCalendar date: 'yyyymmdd'
    """
    if len(value) != 8:
        raise ValueError('Invalid iCalendar date: %r' % value)
    try:
        y, m, d = int(value[0:4]), int(value[4:6]), int(value[6:8])
    except ValueError:
        raise ValueError('Invalid iCalendar date: %r' % value)
    else:
        return datetime.date(y, m, d)


def parse_date_time(value):
    """Parse iCalendar DATE-TIME value.  Returns a datetime instance.

    A simple usage example:

    >>> parse_date_time('20030405T060708')
    datetime.datetime(2003, 4, 5, 6, 7, 8)

    >>> parse_date_time('20030405T060708Z')
    datetime.datetime(2003, 4, 5, 6, 7, 8, tzinfo=<UTC>)

    Examples of invalid arguments:

    >>> parse_date_time('20030405T060708+05:00')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar date-time: '20030405T060708+05:00'

    >>> parse_date_time('20030405T060708A')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar date-time: '20030405T060708A'

    >>> parse_date_time('')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar date-time: ''

    For timezone tests see tests.test_icalendar.TestParseDateTime.

    """
    datetime_rx = re.compile(r'(\d{4})(\d{2})(\d{2})'
                             r'T(\d{2})(\d{2})(\d{2})(Z?)$')
    match = datetime_rx.match(value)
    if match is None:
        raise ValueError('Invalid iCalendar date-time: %r' % value)
    y, m, d, hh, mm, ss, utc = match.groups()
    dt = datetime.datetime(int(y), int(m), int(d),
                           int(hh), int(mm), int(ss))
    if utc:
        return pytz.utc.localize(dt)
    return dt


def parse_duration(value):
    """Parse iCalendar DURATION value.  Returns a timedelta instance.

    >>> parse_duration('+P11D')
    datetime.timedelta(11)
    >>> parse_duration('-P2W')
    datetime.timedelta(-14)
    >>> parse_duration('P1DT2H3M4S')
    datetime.timedelta(1, 7384)
    >>> parse_duration('P1DT2H3M')
    datetime.timedelta(1, 7380)
    >>> parse_duration('P1DT2H')
    datetime.timedelta(1, 7200)
    >>> parse_duration('PT2H')
    datetime.timedelta(0, 7200)
    >>> parse_duration('PT2H3M4S')
    datetime.timedelta(0, 7384)
    >>> parse_duration('PT3M4S')
    datetime.timedelta(0, 184)
    >>> parse_duration('PT22S')
    datetime.timedelta(0, 22)
    >>> parse_duration('')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar duration: ''
    >>> parse_duration('xyzzy')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar duration: 'xyzzy'
    >>> parse_duration('P')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar duration: 'P'
    >>> parse_duration('P1WT2H')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar duration: 'P1WT2H'
    """
    date_part = r'(\d+)D'
    time_part = r'T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+)S)?'
    datetime_part = '(?:%s)?(?:%s)?' % (date_part, time_part)
    weeks_part = r'(\d+)W'
    duration_rx = re.compile(r'([-+]?)P(?:%s|%s)$'
                             % (weeks_part, datetime_part))
    match = duration_rx.match(value)
    if match is None:
        raise ValueError('Invalid iCalendar duration: %r' % value)
    sign, weeks, days, hours, minutes, seconds = match.groups()
    if weeks:
        value = datetime.timedelta(weeks=int(weeks))
    else:
        if (days is None and hours is None
            and minutes is None and seconds is None):
            raise ValueError('Invalid iCalendar duration: %r'
                             % value)
        value = datetime.timedelta(days=int(days or 0),
                                   hours=int(hours or 0),
                                   minutes=int(minutes or 0),
                                   seconds=int(seconds or 0))
    if sign == '-':
        value = -value
    return value


def parse_period(value):
    """Parse iCalendar PERIOD value.  Returns a Period instance.

    >>> p = parse_period('20030405T060708/20030405T060709')
    >>> print repr(p).replace('),', '),\\n      ')
    Period(datetime.datetime(2003, 4, 5, 6, 7, 8),
           datetime.datetime(2003, 4, 5, 6, 7, 9))
    >>> parse_period('20030405T060708/PT1H1M1S')
    Period(datetime.datetime(2003, 4, 5, 6, 7, 8), datetime.timedelta(0, 3661))
    >>> parse_period('xyzzy')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar period: 'xyzzy'
    >>> parse_period('foo/foe')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar period: 'foo/foe'
    """
    parts = value.split('/')
    if len(parts) != 2:
        raise ValueError('Invalid iCalendar period: %r' % value)
    try:
        start = parse_date_time(parts[0])
        try:
            end_or_duration = parse_date_time(parts[1])
        except ValueError:
            end_or_duration = parse_duration(parts[1])
    except ValueError:
        raise ValueError('Invalid iCalendar period: %r' % value)
    else:
        return Period(start, end_or_duration)


def _parse_recurrence_weekly(args):
    """Parse iCalendar weekly recurrence rule.

    args is a mapping from attribute names in RRULE to their string values.

    >>> _parse_recurrence_weekly({})
    WeeklyRecurrenceRule(1, None, None, (), ())
    >>> _parse_recurrence_weekly({'BYDAY': 'WE'})
    WeeklyRecurrenceRule(1, None, None, (), (2,))
    >>> _parse_recurrence_weekly({'BYDAY': 'MO,WE,SU'})
    WeeklyRecurrenceRule(1, None, None, (), (0, 2, 6))

    """
    from schooltool.calendar.recurrent import WeeklyRecurrenceRule
    weekdays = []
    days = args.get('BYDAY', None)
    if days is not None:
        for day in days.split(','):
            weekdays.append(ical_weekdays.index(day))
    return WeeklyRecurrenceRule(weekdays=weekdays)


def _parse_recurrence_monthly(args):
    """Parse iCalendar monthly recurrence rule.

    args is a mapping from attribute names in RRULE to their string values.

    Month-day recurrency is the default:

    >>> _parse_recurrence_monthly({})
    MonthlyRecurrenceRule(1, None, None, (), 'monthday')

    3rd Tuesday in a month:

    >>> _parse_recurrence_monthly({'BYDAY': '3TU'})
    MonthlyRecurrenceRule(1, None, None, (), 'weekday')

    Last Wednesday:

    >>> _parse_recurrence_monthly({'BYDAY': '-1WE'})
    MonthlyRecurrenceRule(1, None, None, (), 'lastweekday')
    """
    from schooltool.calendar.recurrent import MonthlyRecurrenceRule
    if 'BYDAY' in args:
        if args['BYDAY'][0] == '-':
            monthly = 'lastweekday'
        else:
            monthly = 'weekday'
    else:
        monthly = 'monthday'
    return MonthlyRecurrenceRule(monthly=monthly)


def parse_recurrence_rule(value):
    """Parse iCalendar RRULE entry.

    Returns the corresponding subclass of RecurrenceRule.

    params is a mapping from attribute names in RRULE to their string values,

    A trivial example of a daily recurrence:

    >>> parse_recurrence_rule('FREQ=DAILY')
    DailyRecurrenceRule(1, None, None, ())

    A slightly more complex example:

    >>> parse_recurrence_rule('FREQ=DAILY;INTERVAL=5;COUNT=7')
    DailyRecurrenceRule(5, 7, None, ())

    An example that includes use of UNTIL:

    >>> parse_recurrence_rule('FREQ=DAILY;UNTIL=20041008T000000')
    DailyRecurrenceRule(1, None, datetime.date(2004, 10, 8), ())
    >>> parse_recurrence_rule('FREQ=DAILY;UNTIL=20041008')
    DailyRecurrenceRule(1, None, datetime.date(2004, 10, 8), ())

    Of course, other recurrence frequencies may be used:

    >>> parse_recurrence_rule('FREQ=WEEKLY;BYDAY=MO,WE,SU')
    WeeklyRecurrenceRule(1, None, None, (), (0, 2, 6))
    >>> parse_recurrence_rule('FREQ=MONTHLY')
    MonthlyRecurrenceRule(1, None, None, (), 'monthday')
    >>> parse_recurrence_rule('FREQ=YEARLY')
    YearlyRecurrenceRule(1, None, None, ())

    You have to provide a valid recurrence frequency, or you will get an error:

    >>> parse_recurrence_rule('')
    Traceback (most recent call last):
      ...
    ValueError: Invalid frequency of recurrence: None
    >>> parse_recurrence_rule('FREQ=bogus')
    Traceback (most recent call last):
      ...
    ValueError: Invalid frequency of recurrence: 'bogus'

    Unknown keys in params are ignored silently:

    >>> parse_recurrence_rule('FREQ=DAILY;WHATEVER=IGNORED')
    DailyRecurrenceRule(1, None, None, ())

    >>> parse_recurrence_rule('FREQ=MONTHLY;INTERVAL=1;UNTIL=20070102;BYMONTHDAY=3')
    MonthlyRecurrenceRule(1, None, datetime.date(2007, 1, 2), (), 'monthday')

    """
    from schooltool.calendar.recurrent import DailyRecurrenceRule
    from schooltool.calendar.recurrent import YearlyRecurrenceRule

    # split up the given value into parameters
    params = {}
    if value:
        for pair in value.split(';'):
            k, v = pair.split('=', 1)
            params[k] = v

    # parse common recurrency attributes
    interval = int(params.pop('INTERVAL', '1'))
    count = params.pop('COUNT', None)
    if count is not None:
        count = int(count)
    until = params.pop('UNTIL', None)
    if until is not None:
        if len(until) == 8:
            until = parse_date(until)
        else:
            until = parse_date_time(until).date()

    # instantiate the corresponding recurrence rule
    freq = params.pop('FREQ', None)
    if freq == 'DAILY':
        rule = DailyRecurrenceRule()
    elif freq == 'WEEKLY':
        rule = _parse_recurrence_weekly(params)
    elif freq == 'MONTHLY':
        rule = _parse_recurrence_monthly(params)
    elif freq == 'YEARLY':
        rule = YearlyRecurrenceRule()
    else:
        raise ValueError('Invalid frequency of recurrence: %r' % freq)

    return rule.replace(interval=interval, count=count, until=until)


class VEventParser(object):
    """iCalendar event.

    Life cycle: when a VEvent is created, a number of properties should be
    added to it using the add method.  Then validate should be called.
    After that you can start using query methods (getOne, hasProp, iterDates).

    Events are classified into two kinds:
     - normal events
     - all-day events

    All-day events are identified by their DTSTART property having a DATE value
    instead of the default DATE-TIME.  All-day events should satisfy the
    following requirements (otherwise an exception will be raised):
     - DURATION property (if defined) should be an integral number of days
     - DTEND property (if defined) should have a DATE value
     - any RDATE and EXDATE properties should only contain DATE values

    The first two requirements are stated in RFC 2445; I'm not so sure about
    the third one.
    """

    default_type = {
        # Default value types for some properties
        'DTSTAMP': 'DATE-TIME',
        'DTSTART': 'DATE-TIME',
        'CREATED': 'DATE-TIME',
        'DTEND': 'DATE-TIME',
        'DURATION': 'DURATION',
        'LAST-MODIFIED': 'DATE-TIME',
        'PRIORITY': 'INTEGER',
        'RECURRENCE-ID': 'DATE-TIME',
        'SEQUENCE': 'INTEGER',
        'URL': 'URI',
        'ATTACH': 'URI',
        'EXDATE': 'DATE-TIME',
        'EXRULE': 'RECUR',
        'RDATE': 'DATE-TIME',
        'RRULE': 'RECUR',
        'LOCATION': 'TEXT',
        'UID': 'TEXT',
    }

    converters = {
        'INTEGER': int,
        'DATE': parse_date,
        'DATE-TIME': parse_date_time,
        'DURATION': parse_duration,
        'PERIOD': parse_period,
        'TEXT': parse_text,
        'RECUR': parse_recurrence_rule,
    }

    singleton_properties = set([
        'DTSTAMP',
        'DTSTART',
        'UID',
        'CLASS',
        'CREATED',
        'DESCRIPTION',
        'DTEND',
        'DURATION',
        'GEO',
        'LAST-MODIFIED',
        'LOCATION',
        'ORGANIZER',
        'PRIORITY',
        'RECURRENCE-ID',
        'SEQUENCE',
        'STATUS',
        'SUMMARY',
        'TRANSP',
        'URL',
    ])

    rdate_types = set(['DATE', 'DATE-TIME', 'PERIOD'])
    exdate_types = set(['DATE', 'DATE-TIME'])

    def __init__(self):
        self._props = {}

    def add(self, property, value, params=None):
        """Add a property.

        Property name is case insensitive.  Params should be a dict from
        uppercased parameter names to their values.

        Multiple calls to add with the same property name override the value.
        This is sufficient for now, but will have to be changed soon.
        """
        if params is None:
            params = {}
        key = property.upper()
        if key in self._props:
            if key in self.singleton_properties:
                raise ICalParseError("Property %s cannot occur more than once"
                                     % key)
            self._props[key].append((value, params))
        else:
            self._props[key] = [(value, params)]

    def validate(self):
        """Check that this event has all the necessary properties."""
        if not self.hasProp('UID'):
            raise ICalParseError("VEVENT must have a UID property")
        if not self.hasProp('DTSTART'):
            raise ICalParseError("VEVENT must have a DTSTART property")
        if self._getType('DTSTART') not in ('DATE', 'DATE-TIME'):
            raise ICalParseError("DTSTART property should have a DATE or"
                                 " DATE-TIME value")
        if self.hasProp('DTEND'):
            if self._getType('DTEND') != self._getType('DTSTART'):
                raise ICalParseError("DTEND property should have the same type"
                                     " as DTSTART")
        elif self.hasProp('DURATION'):
            if self._getType('DURATION') != 'DURATION':
                raise ICalParseError("DURATION property should have type"
                                     " DURATION")

    def parse(self):
        """Construct the corresponding VEvent object.

        Returns a VEvent.
        """
        self.validate()
        uid = self.getOne('UID')
        summary = self.getOne('SUMMARY')

        all_day_event = self._getType('DTSTART') == 'DATE'

        dtstart = self.getOne('DTSTART')
        prop = self._props['DTSTART'][0]
        dtstart_tzid = prop[1].get('TZID')

        if self.hasProp('DURATION'):
            duration = self.getOne('DURATION')
            dtend = dtstart + duration
            dtend_tzid = dtstart_tzid
        else:
            dtend = self.getOne('DTEND', None)
            if dtend is not None:
                dtend_tzid = self._props['DTEND'][0][1].get('TZID')
            else:
                dtend_tzid = dtstart_tzid
            if dtend is None:
                dtend = dtstart
                if all_day_event:
                    dtend += datetime.date.resolution

        location = self.getOne('LOCATION', None)
        description = self.getOne('DESCRIPTION', None)

        if dtstart > dtend:
            raise ICalParseError("Event start time should precede end time")
        elif all_day_event and dtstart == dtend:
            raise ICalParseError("Event start time should precede end time")

        rdates = self._extractListOfDates('RDATE', self.rdate_types,
                                          all_day_event)
        exdates = self._extractListOfDates('EXDATE', self.exdate_types,
                                           all_day_event)

        rrule = self.getOne('RRULE', None)
        if rrule is not None and exdates:
            exceptions = [datetime.date(dt.year, dt.month, dt.day)
                          for dt in exdates]
            rrule = rrule.replace(exceptions=exceptions)

        return VEvent(uid=uid, summary=summary, all_day_event=all_day_event,
                      dtstart=dtstart, dtstart_tzid=dtstart_tzid, dtend=dtend,
                      dtend_tzid=dtend_tzid, location=location,
                      description=description, rrule=rrule, rdates=rdates,
                      exdates=exdates)

    def _extractListOfDates(self, key, accepted_types, all_day_event):
        """Parse a comma separated list of values.

        If all_day_event is True, only accepts DATE values.  Otherwise accepts
        all value types listed in 'accepted_types'.
        """
        dates = []
        default_type = self.default_type[key]
        for value, params in self._props.get(key, []):
            value_type = params.get('VALUE', default_type)
            if value_type not in accepted_types:
                raise ICalParseError('Invalid value type for %s: %s'
                                     % (key, value_type))
            if all_day_event and value_type != 'DATE':
                raise ICalParseError('I do not understand how to interpret '
                                     '%s values in %s for all-day events.'
                                     % (value_type, key))
            converter = self.converters.get(value_type)
            dates.extend(map(converter, value.split(',')))
        return dates

    def _getType(self, property):
        """Return the type of the property value.

        Only call getType for properties that do not occur more than once.
        """
        key = property.upper()
        values = self._props[key]
        assert len(values) == 1
        value, params = values[0]
        default_type = self.default_type.get(key, 'TEXT')
        return params.get('VALUE', default_type)

    def getOne(self, property, default=None):
        """Return the value of a property as an appropriate Python object.

        Only call getOne for properties that do not occur more than once.
        """
        try:
            values = self._props[property.upper()]
            assert len(values) == 1
            value, params = values[0]
        except KeyError:
            return default
        else:
            converter = self.converters.get(self._getType(property))
            if converter is None:
                return value
            else:
                return converter(value)

    def hasProp(self, property):
        """Return True if this VEvent has a named property."""
        return property.upper() in self._props


class VEvent(object):
    """A calendar event.

    Attributes:

          uid               The unique id of this event
          summary           Textual summary of this event
          all_day_event     True if this is an all-day event
          dtstart           start of the event (inclusive)
          dtstart_tzid      timezone for the start of the event
          dtend             end of the event (not inclusive)
          dtend_tzid        timezone for the end of the event
          location          location of the event
          description       description of the event
          rrule             recurrency rule
          rdates            a list of recurrence dates or periods
          exdates           a list of exception dates

    """

    def __init__(self, uid, summary, all_day_event, dtstart, dtstart_tzid,
                 dtend, dtend_tzid, location, description, rrule, rdates,
                 exdates):
        self.uid = uid
        self.summary = summary
        self.all_day_event = all_day_event
        self.dtstart = dtstart
        self.dtstart_tzid = dtstart_tzid
        self.dtend = dtend
        self.dtend_tzid = dtend_tzid
        self.location = location
        self.description = description
        self.rrule = rrule
        self.rdates = rdates
        self.exdates = exdates

    @staticmethod
    def parse(rows):
        """Parse a list of rows into a VEvent object."""
        parser = VEventParser()
        props, blocks = parse_block(rows[1:-1])

        for key, value, params in props:
            parser.add(key, value, params)

        return parser.parse()


class Period(object):
    """A period of time."""

    def __init__(self, start, end_or_duration):
        self.start = start
        self.end_or_duration = end_or_duration
        if isinstance(end_or_duration, datetime.timedelta):
            self.duration = end_or_duration
            self.end = self.start + self.duration
        else:
            self.end = end_or_duration
            self.duration = self.end - self.start
        if self.start > self.end:
            raise ValueError("Start time is greater than end time")

    def __repr__(self):
        return "Period(%r, %r)" % (self.start, self.end_or_duration)

    def __cmp__(self, other):
        if not isinstance(other, Period):
            raise NotImplementedError('Cannot compare Period with %r' % other)
        return cmp((self.start, self.end), (other.start, other.end))

    def overlaps(self, other):
        if self.start > other.start:
            return other.overlaps(self)
        if self.start <= other.start < self.end:
            return True
        return False


class ICalParseError(Exception):
    """Invalid syntax in an iCalendar file."""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Functional tests for schooltool.email

"""
import smtplib
import socket

from schooltool.email.mail import EmailUtility


class StubConnection(object):

    reject_mail_to = 'reject.com'

    host = None
    port = None

    def connect(self, host, port):
        self.host, self.port = host, port
        if port == '255':
            raise socket.error('StubConnection failed successfuly')

    def ehlo(self):
        return 200, 'okay'

    def helo(self):
        return 200, 'okay'

    def has_extn(self, extention):
        return True

    @property
    def does_esmtp(self):
        return True

    def login(self, username, password):
        if self.host == 'fail_login':
            raise smtplib.SMTPException('Purposefully failed to log in')

    def quit(self):
        pass

    def sendmail(self, from_address, to_addresses, message):
        if self.reject_mail_to in from_address:
            raise smtplib.SMTPSenderRefused(
                -1, 'Failed successfuly', from_address)
        rejected_recipients = dict([
            (addr, 'Fail') for addr in to_addresses
            if self.reject_mail_to in addr])
        if rejected_recipients:
            if len(rejected_recipients) == to_addresses:
                raise smtplib.SMTPRecipientsRefused(rejected_recipients)
            else:
                return rejected_recipients
        if self.host == 'reject_malformed':
            raise smtplib.SMTPDataError(-1, 'I pretend that this is bad data')


class SentMessages(list):
    """This is basically a list of email messages with a helper to
    ease writing of functional tests.
    """

    def print_mail(self, emails=None):
        """Pass an email or list of emails to print.
        Alternatively don't pass anything to print them all.
        """
        if emails is None:
            emails = self
        elif not hasattr(emails, '__iter__'):
            emails = [emails]

        for email in emails:
            headers = [
                'From: %s' % email.from_address,
                'To: %s' % ', '.join(email.to_addresses),
                'Subject: %s' % email.subject]
            maxlen = max(
                [len(s) for s in headers + email.body.split('\n')])
            print '\n'.join(
                ['=' * maxlen] +
                headers +
                ['-' * maxlen,
                 '%s' % email.body]
                )


class StubEmailUtility(EmailUtility):
    """Implementation of IEmailUtility that fails in variety of ways depending
    on the contents of email being sent and logs the messages that are
    successfuly sent.
    """

    smtp_factory = StubConnection

    def __init__(self, *args, **kw):
        super(StubEmailUtility, self).__init__(*args, **kw)
        self.sent = SentMessages()

    def send(self, email):
        result = EmailUtility.send(self, email)
        if result:
            self.sent.append(email)
        return result



##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Bootstrap a buildout-based project

Simply run this script in a directory containing a buildout.cfg.
The script accepts buildout command-line options, so you can
use the -c option to specify an alternate configuration file.
"""

import os, shutil, sys, tempfile, urllib, urllib2, subprocess
from optparse import OptionParser

if sys.platform == 'win32':
    def quote(c):
        if ' ' in c:
            return '"%s"' % c  # work around spawn lamosity on windows
        else:
            return c
else:
    quote = str

# See zc.buildout.easy_install._has_broken_dash_S for motivation and comments.
stdout, stderr = subprocess.Popen(
    [sys.executable, '-Sc',
     'try:\n'
     '    import ConfigParser\n'
     'except ImportError:\n'
     '    print 1\n'
     'else:\n'
     '    print 0\n'],
    stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
has_broken_dash_S = bool(int(stdout.strip()))

# In order to be more robust in the face of system Pythons, we want to
# run without site-packages loaded.  This is somewhat tricky, in
# particular because Python 2.6's distutils imports site, so starting
# with the -S flag is not sufficient.  However, we'll start with that:
if not has_broken_dash_S and 'site' in sys.modules:
    # We will restart with python -S.
    args = sys.argv[:]
    args[0:0] = [sys.executable, '-S']
    args = map(quote, args)
    os.execv(sys.executable, args)
# Now we are running with -S.  We'll get the clean sys.path, import site
# because distutils will do it later, and then reset the path and clean
# out any namespace packages from site-packages that might have been
# loaded by .pth files.
clean_path = sys.path[:]
import site  # imported because of its side effects
sys.path[:] = clean_path
for k, v in sys.modules.items():
    if k in ('setuptools', 'pkg_resources') or (
        hasattr(v, '__path__') and
        len(v.__path__) == 1 and
        not os.path.exists(os.path.join(v.__path__[0], '__init__.py'))):
        # This is a namespace package.  Remove it.
        sys.modules.pop(k)

is_jython = sys.platform.startswith('java')

setuptools_source = 'http://peak.telecommunity.com/dist/ez_setup.py'
distribute_source = 'http://python-distribute.org/distribute_setup.py'


# parsing arguments
def normalize_to_url(option, opt_str, value, parser):
    if value:
        if '://' not in value:  # It doesn't smell like a URL.
            value = 'file://%s' % (
                urllib.pathname2url(
                    os.path.abspath(os.path.expanduser(value))),)
        if opt_str == '--download-base' and not value.endswith('/'):
            # Download base needs a trailing slash to make the world happy.
            value += '/'
    else:
        value = None
    name = opt_str[2:].replace('-', '_')
    setattr(parser.values, name, value)

usage = '''\
[DESIRED PYTHON FOR BUILDOUT] bootstrap.py [options]

Bootstraps a buildout-based project.

Simply run this script in a directory containing a buildout.cfg, using the
Python that you want bin/buildout to use.

Note that by using --setup-source and --download-base to point to
local resources, you can keep this script from going over the network.
'''

parser = OptionParser(usage=usage)
parser.add_option("-v", "--version", dest="version",
                          help="use a specific zc.buildout version")
parser.add_option("-d", "--distribute",
                   action="store_true", dest="use_distribute", default=False,
                   help="Use Distribute rather than Setuptools.")
parser.add_option("--setup-source", action="callback", dest="setup_source",
                  callback=normalize_to_url, nargs=1, type="string",
                  help=("Specify a URL or file location for the setup file. "
                        "If you use Setuptools, this will default to " +
                        setuptools_source + "; if you use Distribute, this "
                        "will default to " + distribute_source + "."))
parser.add_option("--download-base", action="callback", dest="download_base",
                  callback=normalize_to_url, nargs=1, type="string",
                  help=("Specify a URL or directory for downloading "
                        "zc.buildout and either Setuptools or Distribute. "
                        "Defaults to PyPI."))
parser.add_option("--eggs",
                  help=("Specify a directory for storing eggs.  Defaults to "
                        "a temporary directory that is deleted when the "
                        "bootstrap script completes."))
parser.add_option("-t", "--accept-buildout-test-releases",
                  dest='accept_buildout_test_releases',
                  action="store_true", default=False,
                  help=("Normally, if you do not specify a --version, the "
                        "bootstrap script and buildout gets the newest "
                        "*final* versions of zc.buildout and its recipes and "
                        "extensions for you.  If you use this flag, "
                        "bootstrap and buildout will get the newest releases "
                        "even if they are alphas or betas."))
parser.add_option("-c", None, action="store", dest="config_file",
                   help=("Specify the path to the buildout configuration "
                         "file to be used."))

options, args = parser.parse_args()

if options.eggs:
    eggs_dir = os.path.abspath(os.path.expanduser(options.eggs))
else:
    eggs_dir = tempfile.mkdtemp()

if options.setup_source is None:
    if options.use_distribute:
        options.setup_source = distribute_source
    else:
        options.setup_source = setuptools_source

if options.accept_buildout_test_releases:
    args.insert(0, 'buildout:accept-buildout-test-releases=true')

try:
    import pkg_resources
    import setuptools  # A flag.  Sometimes pkg_resources is installed alone.
    if not hasattr(pkg_resources, '_distribute'):
        raise ImportError
except ImportError:
    ez_code = urllib2.urlopen(
        options.setup_source).read().replace('\r\n', '\n')
    ez = {}
    exec ez_code in ez
    setup_args = dict(to_dir=eggs_dir, download_delay=0)
    if options.download_base:
        setup_args['download_base'] = options.download_base
    if options.use_distribute:
        setup_args['no_fake'] = True
        if sys.version_info[:2] == (2, 4):
            setup_args['version'] = '0.6.32'
    ez['use_setuptools'](**setup_args)
    if 'pkg_resources' in sys.modules:
        reload(sys.modules['pkg_resources'])
    import pkg_resources
    # This does not (always?) update the default working set.  We will
    # do it.
    for path in sys.path:
        if path not in pkg_resources.working_set.entries:
            pkg_resources.working_set.add_entry(path)

cmd = [quote(sys.executable),
       '-c',
       quote('from setuptools.command.easy_install import main; main()'),
       '-mqNxd',
       quote(eggs_dir)]

if not has_broken_dash_S:
    cmd.insert(1, '-S')

find_links = options.download_base
if not find_links:
    find_links = os.environ.get('bootstrap-testing-find-links')
if not find_links and options.accept_buildout_test_releases:
    find_links = 'http://downloads.buildout.org/'
if find_links:
    cmd.extend(['-f', quote(find_links)])

if options.use_distribute:
    setup_requirement = 'distribute'
else:
    setup_requirement = 'setuptools'
ws = pkg_resources.working_set
setup_requirement_path = ws.find(
    pkg_resources.Requirement.parse(setup_requirement)).location
env = dict(
    os.environ,
    PYTHONPATH=setup_requirement_path)

requirement = 'zc.buildout'
version = options.version
if version is None and not options.accept_buildout_test_releases:
    # Figure out the most recent final version of zc.buildout.
    import setuptools.package_index
    _final_parts = '*final-', '*final'

    def _final_version(parsed_version):
        for part in parsed_version:
            if (part[:1] == '*') and (part not in _final_parts):
                return False
        return True
    index = setuptools.package_index.PackageIndex(
        search_path=[setup_requirement_path])
    if find_links:
        index.add_find_links((find_links,))
    req = pkg_resources.Requirement.parse(requirement)
    if index.obtain(req) is not None:
        best = []
        bestv = None
        for dist in index[req.project_name]:
            distv = dist.parsed_version
            if distv >= pkg_resources.parse_version('2dev'):
                continue
            if _final_version(distv):
                if bestv is None or distv > bestv:
                    best = [dist]
                    bestv = distv
                elif distv == bestv:
                    best.append(dist)
        if best:
            best.sort()
            version = best[-1].version

if version:
    requirement += '=='+version
else:
    requirement += '<2dev'

cmd.append(requirement)

if is_jython:
    import subprocess
    exitcode = subprocess.Popen(cmd, env=env).wait()
else:  # Windows prefers this, apparently; otherwise we would prefer subprocess
    exitcode = os.spawnle(*([os.P_WAIT, sys.executable] + cmd + [env]))
if exitcode != 0:
    sys.stdout.flush()
    sys.stderr.flush()
    print ("An error occurred when trying to install zc.buildout. "
           "Look above this message for any errors that "
           "were output by easy_install.")
    sys.exit(exitcode)

ws.add_entry(eggs_dir)
ws.require(requirement)
import zc.buildout.buildout

# If there isn't already a command in the args, add bootstrap
if not [a for a in args if '=' not in a]:
    args.append('bootstrap')


# if -c was provided, we push it back into args for buildout's main function
if options.config_file is not None:
    args[0:0] = ['-c', options.config_file]

zc.buildout.buildout.main(args)
if not options.eggs:  # clean up temporary egg directory
    shutil.rmtree(eggs_dir)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool metaconfiguration code.
"""

from zope.interface import implements
from zope.component import provideAdapter, provideSubscriptionAdapter
from zope.component import queryUtility, getGlobalSiteManager
from zope.container.btree import BTreeContainer

from schooltool.securitypolicy.crowds import AggregateCrowd
from schooltool.securitypolicy.interfaces import ICrowd
from schooltool.securitypolicy.interfaces import ICrowdToDescribe
from schooltool.securitypolicy.interfaces import ICrowdDescription
from schooltool.securitypolicy.interfaces import IAccessControlSetting
from schooltool.securitypolicy.interfaces import IAccessControlCustomisations
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.securitypolicy.crowds import getCrowdsUtility
from schooltool.securitypolicy.crowds import getDescriptionUtility
from schooltool.securitypolicy.crowds import DescriptionGroup
from schooltool.securitypolicy.crowds import GroupAction
from schooltool.securitypolicy.crowds import CrowdDescription


# ZCML execution order
ZCML_REGISTER_DESCRIPTION_GROUPS       = 10100
ZCML_REGISTER_DESCRIPTION_ACTIONS      = 10200
ZCML_REGISTER_CROWD_DESCRIPTIONS       = 10300
ZCML_REGISTER_DESCRIPTION_SWITCHING    = 10400


class AggregateUtilityCrowd(AggregateCrowd):
    interface = None
    permission = None

    def crowdFactories(self):
        return getCrowdsUtility().getFactories(self.permission, self.interface)


def registerCrowdAdapter(permission, interface):
    """Register an adapter to ICrowd for interface.

    The adapter dynamically retrieves the list of crowds from the
    global objcrowds.  You should not call this function several times
    for the same (permission, interface).
    """
    aggregator_class = type(
        str('%s_%s' % (AggregateUtilityCrowd.__name__, interface.__name__)),
        (AggregateUtilityCrowd, ),
        {'interface': interface, 'permission': permission})

    provideAdapter(aggregator_class, provides=ICrowd, adapts=[interface],
                   name=permission)


def handle_crowd(name, factory):
    """Handler for the ZCML <crowd> directive."""
    getCrowdsUtility().factories[name] = factory


def handle_allow(crowdname, permission, interface):
    """Handler for the ZCML <allow> directive.

    interface is the interface for which the security declaration is issued,
    crowdname is a string,
    permission is an identifier for a permission.

    The function registers the given crowd factory in the ICrowdsUtility.

    An adapter to ICrowd is provided if interface is specified.
    """

    utility = getCrowdsUtility()

    discriminator = (permission, interface)
    if discriminator not in utility.crowds:
        utility.crowds[discriminator] = []
        if interface is not None:
            registerCrowdAdapter(permission, interface)

    utility.crowds[discriminator].append(crowdname)


def crowd(_context, name, factory):
    # Declare the crowd.
    _context.action(discriminator=('crowd', name), callable=handle_crowd,
                    args=(name, factory))


def allow(_context, interface=None, crowds=None, permission=None):
    for crowd in crowds:
        _context.action(discriminator=('allow', crowd, permission, interface),
                        callable=handle_allow,
                        args=(crowd, permission, interface))


def deny(_context, interface=None, crowds=None, permission=None):
    # XXX: Deny directive needs documentation.
    for crowd in crowds:
        _context.action(discriminator=('allow', crowd, permission, interface),
                        callable=lambda: None,
                        args=())


class AccessControlSetting(object):
    implements(IAccessControlSetting)

    def __init__(self, key, text, alt_text, default):
        self.key = key
        self.text = text
        self.alt_text = alt_text
        self.default = default

    def getValue(self):
        app = ISchoolToolApplication(None)
        customisations = IAccessControlCustomisations(app)
        return customisations.get(self.key)

    def setValue(self, value):
        app = ISchoolToolApplication(None)
        customisations = IAccessControlCustomisations(app)
        return customisations.set(self.key, value)

    def __repr__(self):
        if self.alt_text is None:
            return "<AccessControlSetting key=%s, text=%s, default=%s>" % (
                self.key, self.text, self.default)
        return "<AccessControlSetting key=%s, text=%s, alt_text=%s, default=%s>" % (
            self.key, self.text, self.alt_text, self.default)


def handle_setting(key, text, alt_text, default):
    def accessControlSettingFactory(context=None):
        return AccessControlSetting(key, text, alt_text, default)
    provideSubscriptionAdapter(accessControlSettingFactory,
                               adapts=[None],
                               provides=IAccessControlSetting)

def setting(_context, key=None, text=None, alt_text=None, default=None):
    _context.action(
        discriminator=('setting', key),
        callable=handle_setting, args=(key, text, alt_text, default))


def handle_aggregate_crowd(name, crowd_names):
    factories = getCrowdsUtility().factories
    try:
        crowds = [factories[crowd_name] for crowd_name in crowd_names]
    except KeyError:
        raise ValueError("invalid crowd id", crowd_name)

    class AggregateCrowdFactory(AggregateCrowd):
        def crowdFactories(self):
            return crowds
    handle_crowd(name, AggregateCrowdFactory)


def aggregate_crowd(_context, name, crowds):
    _context.action(discriminator=('crowd', name),
                    callable=handle_aggregate_crowd, args=(name, crowds))


def handle_group(name, group):
    util = getDescriptionUtility()
    util.groups[name] = group
    group.__name__ = name
    group.__parent__ = util.groups


def describe_group(_context, name=None, title=None, description=None, klass=None):
    if klass is None and title is None:
        raise TypeError("Must specify title or klass.")

    if klass is None:
        klass = DescriptionGroup

    new_class = type(str('%s_%s' % (klass.__name__, name.capitalize())),
                     (klass, ), {})
    group = new_class()
    if title is not None:
        group.title = title
    if description is not None:
        group.description = description

    _context.action(discriminator=('describe_group', name),
                    callable=handle_group,
                    args=(name, group),
                    order=ZCML_REGISTER_DESCRIPTION_GROUPS)


def handle_action(group_name, name, action):
    util = getDescriptionUtility()
    if group_name not in util.actions_by_group:
        util.actions_by_group[group_name]=BTreeContainer()
    util.actions_by_group[group_name][name] = action
    action.__name__ = name
    action.__parent__ = util.actions_by_group[group_name]


def describe_action(_context, group=None, name=None, order=0,
                    interface=None, permission='',
                    title=None, description=None, klass=None):
    if klass is None and title is None:
        raise TypeError("Must specify title or klass.")

    if klass is None:
        klass = GroupAction

    new_class = type(str('%s_%s' % (klass.__name__, name.capitalize())),
                     (klass, ), {})
    action = new_class()
    if title is not None:
        action.title = title
    if description is not None:
        action.description = description
    if interface is not None:
        action.interface = interface
    if permission is not None:
        action.permission = permission
    action.order = order

    _context.action(discriminator=('describe_action', group, name),
                    callable=handle_action,
                    args=(group, name, action),
                    order=ZCML_REGISTER_DESCRIPTION_ACTIONS)


def handle_crowd_description(context, group_id, action_id,
                             interface, permission_id,
                             crowd_getter, description_factory):
    util = getDescriptionUtility()

    group = None
    if group_id is not None:
        group = util.groups[group_id]

    action = None
    if action_id is not None:
        action = util.actions_by_group[group_id][action_id]

    crowd_factory = crowd_getter()

    provideAdapter(description_factory,
                   provides=ICrowdDescription,
                   adapts=[crowd_factory,
                           action and action.__class__ or None,
                           group and group.__class__ or None,
                           ])


def describe_crowd(_context, group=None, action=None,
                   interface=None, permission=None,
                   crowd=None, crowd_factory=None,
                   factory=None, title=None, description=None):

    if crowd is not None and crowd_factory is not None:
        raise TypeError("Must specify either crowd or crowd_factory.")

    if action is not None and group is None:
        raise TypeError("Must specify group when specifying action.")

    if crowd is not None:
        crowd_getter = lambda: getCrowdsUtility().getFactory(crowd)
    elif crowd_factory is not None:
        crowd_getter = lambda: crowd_factory
    else:
        crowd_getter = lambda: None

    if factory is None:
        if title is None and description is None:
            raise TypeError("Must specify either description factory"
                            " or title/description.")
        factory = CrowdDescription

    factory_dict = {}
    if title is not None:
        factory_dict['title'] = title
    if description is not None:
        factory_dict['description'] = description
    new_factory = type(factory.__name__, (factory, ), factory_dict)

    discriminator = ('describe_crowd', group, action,
                     interface, permission, crowd, crowd_factory)

    _context.action(discriminator=discriminator,
                    callable=handle_crowd_description,
                    args=(_context, group, action,
                          interface, permission, crowd_getter,
                          new_factory),
                    order=ZCML_REGISTER_CROWD_DESCRIPTIONS)


def handle_switch_description(context, group_id, action_id,
                              crowd_getter, replacement_crowd_getter):
    util = getDescriptionUtility()

    group = None
    if group_id is not None:
        group = util.groups[group_id]

    action = None
    if action_id is not None:
        action = util.actions_by_group[group_id][action_id]

    replacement = replacement_crowd_getter()
    unpack_replacement = lambda crowd, action, group: replacement(crowd)

    provideAdapter(unpack_replacement,
                   provides=ICrowdToDescribe,
                   adapts=[crowd_getter(),
                           action and action.__class__ or None,
                           group and group.__class__ or None,
                           ])


def switch_description(_context,
                       group=None, action=None,
                       crowd=None, crowd_factory=None,
                       use_crowd=None, use_crowd_factory=None):

    if crowd is not None and crowd_factory is not None:
        raise TypeError("Must specify either crowd or crowd_factory.")

    if use_crowd is not None and use_crowd_factory is not None:
        raise TypeError("Must specify either use_crowd or use_crowd_factory.")

    if action is not None and group is None:
        raise TypeError("Must specify group when specifying action.")

    if crowd is not None:
        crowd_getter = lambda: getCrowdsUtility().getFactory(crowd)
    elif crowd_factory is not None:
        crowd_getter = lambda: crowd_factory
    else:
        crowd_getter = lambda: None

    if use_crowd is not None:
        replacement_crowd_getter = lambda: getCrowdsUtility().getFactory(use_crowd)
    elif use_crowd_factory is not None:
        replacement_crowd_getter = lambda: use_crowd_factory
    else:
        replacement_crowd_getter = lambda: None

    discriminator = ('switch_description', group, action,
                     crowd, crowd_factory, use_crowd, use_crowd_factory)

    _context.action(discriminator=discriminator,
                    callable=handle_switch_description,
                    args=(_context, group, action,
                          crowd_getter, replacement_crowd_getter),
                    order=ZCML_REGISTER_DESCRIPTION_SWITCHING)




import stesting

def registerTestSetup():
    from schooltool.testing import registry

    def addSchoolYearContainer(app):
        from schooltool.schoolyear.schoolyear import SchoolYearContainer
        from schooltool.schoolyear.schoolyear import SCHOOLYEAR_CONTAINER_KEY
        app[SCHOOLYEAR_CONTAINER_KEY] = SchoolYearContainer()

    registry.register('ApplicationContainers', addSchoolYearContainer)

registerTestSetup()
del registerTestSetup



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Interfaces for SchoolTool calendar browser views.
"""

import zope.schema
from zope.interface import Interface

from schooltool.calendar.interfaces import ICalendar, ICalendarEvent
from schooltool.common import IDateRange


class IEventForDisplay(ICalendarEvent):
    """A decorated calendar event.

    Calendar views have additional information that is not contained in event
    objects.
    """

    context = zope.schema.Object(
        title=u"The event being decorated",
        schema=ICalendarEvent)

    dtend = zope.schema.Datetime(
        title=u"End",
        readonly=True,
        description=u"""
        Date and time when the event ends.

        dtend == dtstart + duration
        """)

    dtstarttz = zope.schema.Datetime(
        title=u"Start (in user's timezone)",
        readonly=True,
        description=u"""
        Date and time when the event starts, converted to the user's preferred
        timezone.
        """)

    dtendtz = zope.schema.Datetime(
        title=u"End (in user's timezone)",
        readonly=True,
        description=u"""
        Date and time when the event ends, converted to the user's preferred
        timezone.
        """)

    source = zope.schema.Object(
        title=u"Source calendar",
        schema=ICalendar,
        description=u"""
        The calendar the event came from.  A view may display events from
        several overlaid calendars.
        """)

    parent_view_link = zope.schema.TextLine(
        title=u"View link",
        description=u"""Link to the view which displays this event""")

    color1 = zope.schema.TextLine(
        title=u"Color used for display (1)",
        description=u"""
        One of the two colors used to distinguish events coming form
        sources.

        These two colors should contrast enough so that one can be used
        as text color when the other is used as a background color.

        This attribute contains a string usable as a CSS color value.
        """)

    color2 = zope.schema.TextLine(
        title=u"Color used for display (2)",
        description=u"""
        The other of the two colors used to distinguish events coming form
        sources.

        These two colors should contrast enough so that one can be used
        as text color when the other is used as a background color.

        This attribute contains a string usable as a CSS color value.
        """)

    shortTitle = zope.schema.TextLine(
        title=u"Short Title",
        description=u"""The title of the event, ellipsized if necessary.""")

    cssClass = zope.schema.TextLine(
        title=u"CSS Class",
        description=u"""
        Suggested event class name.  Currently it is always 'event'.
        """)


class IHaveEventLegend(Interface):
    """Classes implementing this interface will have an event legend showing."""

    cursor = zope.schema.Date(
        title=u"Cursor",
        readonly=True,
        description=u"Current date",
        required=False)

    cursor_range = zope.schema.Object(
        title=u"Dates the view spans.",
        readonly=True,
        schema=IDateRange,
        required=False)



#!/usr/bin/env python
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003-2013 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool setup script.
"""

import os
from setuptools import setup, find_packages

if os.path.exists("version.txt"):
    version = open("version.txt").read().strip()
else:
    version = open("version.txt.in").read().strip()

setup(
    name="schooltool",
    description="A common information systems platform for school administration.",
    long_description="""
SchoolTool is an open source school management information system.  It is
a web application, usable with an ordinary browser.

Any modern web browser is suitable for the web application interface.  The
interface degrades gracefully, so a browser that does not support CSS or
Javascript will be usable, although perhaps not very nice or convenient.""",
    version=version,
    url='http://www.schooltool.org',
    license="GPL",
    maintainer="SchoolTool Developers",
    maintainer_email="schooltool-developers@lists.launchpad.net",
    platforms=["any"],
    classifiers=["Development Status :: 5 - Production/Stable",
    "Environment :: Web Environment",
    "Intended Audience :: End Users/Desktop",
    "License :: OSI Approved :: GNU General Public License (GPL)",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Zope",
    "Topic :: Education"],
    package_dir={'': 'src'},
    packages=find_packages('src'),
    namespace_packages=["schooltool"],
    install_requires=['celery>=2.5',
                      'redis',
                      'bottle',
                      'PasteDeploy',
                      'PasteScript',
                      'Pillow',
                      'pytz',
                      'reportlab',
                      'setuptools',
                      'xlrd',
                      'xlwt',
                      'z3c.autoinclude',
                      'z3c.form>=2.0',
                      'z3c.formui',
                      'z3c.macro',
                      'z3c.rml',
                      'z3c.template',
                      'zc.catalog',
                      'zc.datetimewidget',
                      'zc.resourcelibrary',
                      'zc.table',
                      'ZODB3',
                      'zope.annotation',
                      'zope.authentication',
                      'zope.app.applicationcontrol',
                      'zope.app.appsetup',
                      'zope.app.basicskin',
                      'zope.app.broken',
                      'zope.app.component',
                      'zope.app.container',
                      'zope.app.dependable',
                      'zope.app.error',
                      'zope.app.exception',
                      'zope.app.form',
                      'zope.app.generations>=3.5',
                      'zope.app.http',
                      'zope.app.locales',
                      'zope.app.principalannotation',
                      'zope.app.publication',
                      'zope.app.schema',
                      'zope.app.security',
                      'zope.app.wsgi',
                      'zope.browser',
                      'zope.browsermenu',
                      'zope.browserpage>=3.10.1',
                      'zope.cachedescriptors',
                      'zope.catalog',
                      'zope.component>=3.8',
                      'zope.configuration',
                      'zope.container>=3.7.2',
                      'zope.contentprovider',
                      'zope.dublincore>=3.7',
                      'zope.event',
                      'zope.file',
                      'zope.filerepresentation',
                      'zope.formlib>=4.0',
                      'zope.html',
                      'zope.i18n>=3.5',
                      'zope.i18nmessageid',
                      'zope.interface',
                      'zope.intid',
                      'zope.keyreference',
                      'zope.lifecycleevent',
                      'zope.location',
                      'zope.login',
                      'zope.mimetype',
                      'zope.pagetemplate>=3.5',
                      'zope.password',
                      'zope.proxy',
                      'zope.publisher>=3.6',
                      'zope.schema',
                      'zope.security',
                      'zope.securitypolicy',
                      'zope.server',
                      'zope.session',
                      'zope.site',
                      'zope.size',
                      'zope.tales',
                      'zope.testing',
                      'zope.testbrowser',   # XXX for selenium extensions
                      'zope.app.testing',   # XXX to get zope.testbrowser.testing.Browser
                      'zope.traversing>=3.13',
                      'zope.ucol',
                      'zope.viewlet'],
    extras_require={'test': ['lxml',
                             'zope.app.testing',
                             'zope.copypastemove',
                             'zope.exceptions',
                             'zope.principalregistry',
                             'zope.testbrowser',
                             'z3c.form>=2.6',
                             'schooltool.devtools>=0.7.1',
                             'selenium'],
                    'docs': ['Sphinx',
                             'z3c.recipe.sphinxdoc'],
                   },
    include_package_data=True,
    zip_safe=False,
    entry_points = """
    [paste.app_factory]
    main = schooltool.paste.main:schooltool_app_factory
    task_results = schooltool.paste.main:task_result_app_factory

    [console_scripts]
    start-schooltool-instance = schooltool.paste.run:main
    make-schooltool-instance = schooltool.paste.instance:make_schooltool_instance
    schooltool-server = schooltool.app.main:main

    [paste.paster_create_template]
    schooltool_deploy = schooltool.paste.templates:SchoolToolDeploy

    [schooltool.instance_type]
    old = schooltool.standard
    schooltool = schooltool.skin.flourish.instance

    [schooltool.tasks]
    schooltool.export.importer = schooltool.export.importer

    """,
    )



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for SchoolTool XLS export views.
"""
import unittest
import doctest
from datetime import date, time

from schooltool.basicperson.interfaces import IDemographics
from schooltool.basicperson.person import BasicPerson
from schooltool.person.person import Person
from schooltool.export.export import MegaExporter, merge_date_ranges
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.testing import format_table
from schooltool.term.term import Term
from schooltool.common import DateRange
from schooltool.contact.contact import Contact, ContactPersonInfo
from schooltool.contact.interfaces import IContact, IContactable
from schooltool.contact.interfaces import IContactContainer
from schooltool.course.course import Course
from schooltool.course.section import Section
from schooltool.course.interfaces import ICourseContainer
from schooltool.course.interfaces import ISectionContainer
from schooltool.schoolyear.schoolyear import SchoolYear
from schooltool.schoolyear.testing import provideStubUtility
from schooltool.schoolyear.testing import provideStubAdapter
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.export.ftesting import export_functional_layer
from schooltool.resource.resource import Location, Equipment
from schooltool.schoolyear.testing import setUp
from schooltool.schoolyear.testing import tearDown
from schooltool.timetable.interfaces import IScheduleContainer
from schooltool.timetable.interfaces import ITimetableContainer
from schooltool.timetable.timetable import SelectedPeriodsSchedule
from schooltool.timetable.tests.test_timetable import TimetableForTests


def setUpSchool(app):
    sy = ISchoolYearContainer(app)['2005'] = SchoolYear('2005',
                                                        date(2005, 1, 1),
                                                        date(2005, 1, 30))

    term = sy['spring'] = Term('Spring', date(2005, 1, 1),
                                         date(2005, 1, 30))
    term.addWeekdays(0, 1, 2, 3, 4)

    pc = app['persons']
    teacher = pc['teacher'] = BasicPerson("teacher", "Mister", "T")
    s1 = pc['john'] = BasicPerson("john", "John", "Peterson")
    s2 = pc['pete'] = BasicPerson("pete", "Pete", "Johnson")

    contacts = IContactContainer(app)
    contact = Contact()
    contact.__name__ = 'pete_parent'
    contact.prefix = 'Ms.'
    contact.first_name = 'Susan'
    contact.middle_name = 'T.'
    contact.last_name = 'Johnson'
    contact.suffix = 'Jr.'
    contact.address_line_1 = '1 First St.'
    contact.address_line_2 = 'Apt. 1'
    contact.city = 'NY'
    contact.state = 'NY'
    contact.country = 'USA'
    contact.postal_code = '00000'
    contact.email = 'davejohnson@gmail.com'
    contact.home_phone = '000-0000'
    contact.work_phone = '111-1111'
    contact.mobile_phone = '222-2222'
    contact.language = 'English'
    contacts['pete_parent'] = contact

    info = ContactPersonInfo()
    info.__parent__ = s2
    info.relationship = 'parent'
    IContactable(s2).contacts.add(contact, info)

    info = ContactPersonInfo()
    info.__parent__ = s2
    info.relationship = 'parent'
    IContactable(s2).contacts.add(IContact(teacher), info)

    d1 = IDemographics(s1)
    d1['ID'] = "112323"
    d1['ethnicity'] = u'Asian'
    d1['language'] = "English"
    d1['placeofbirth'] = "Humptington"
    d1['citizenship'] = "US"
    d2 = IDemographics(s2)
    d2['ID'] = "333655"
    d2['ethnicity'] = u'White'
    d2['language'] = "Brittish"
    d2['placeofbirth'] = "Providence"
    d2['citizenship'] = "UK"

    course = ICourseContainer(sy)['c1'] = Course("History")


def doctest_format_school_years():
    """
        >>> app = ISchoolToolApplication(None)
        >>> setUpSchool(app)
        >>> exporter = MegaExporter(app, None)
        >>> for row in exporter.format_school_years(): print row
        [Header('ID'), Header('Title'), Header('Start'), Header('End')]
        [Text(u'2005'), Text('2005'), Date(datetime.date(2005, 1, 1)), Date(datetime.date(2005, 1, 30))]
    """


def doctest_format_terms():
    """
        >>> app = ISchoolToolApplication(None)
        >>> setUpSchool(app)
        >>> exporter = MegaExporter(app, None)
        >>> for row in exporter.format_terms(): print row
        [Header('SchoolYear'), Header('ID'), Header('Title'), Header('Start'), Header('End')]
        [Text(u'2005'), Text(u'spring'), Text('Spring'), Date(datetime.date(2005, 1, 1)), Date(datetime.date(2005, 1, 30))]
        []
        [Header('Weekends')]
        [Text('Monday'), Text('Tuesday'), Text('Wednesday'), Text('Thursday'), Text('Friday'), Text('Saturday'), Text('Sunday')]
        [Text(''), Text(''), Text(''), Text(''), Text(''), Text('X'), Text('X')]

    Let's add some holdays:

        >>> term = ISchoolYearContainer(app)['2005']['spring']
        >>> for day in DateRange(date(2005, 1, 10), date(2005, 1, 17)):
        ...     if term.isSchoolday(day): term.remove(day)

        >>> for row in exporter.format_terms(): print row
        [Header('SchoolYear'), Header('ID'), Header('Title'), Header('Start'), Header('End')]
        [Text(u'2005'), Text(u'spring'), Text('Spring'), Date(datetime.date(2005, 1, 1)), Date(datetime.date(2005, 1, 30))]
        []
        [Header('Holidays')]
        [Date(datetime.date(2005, 1, 10)), Date(datetime.date(2005, 1, 14))]
        [Date(datetime.date(2005, 1, 17)), Date(datetime.date(2005, 1, 17))]
        []
        [Header('Weekends')]
        [Text('Monday'), Text('Tuesday'), Text('Wednesday'), Text('Thursday'), Text('Friday'), Text('Saturday'), Text('Sunday')]
        [Text(''), Text(''), Text(''), Text(''), Text(''), Text('X'), Text('X')]

    And a working weekend:

        >>> for day in DateRange(date(2005, 1, 8), date(2005, 1, 9)):
        ...     term.add(day)

        >>> for row in exporter.format_terms(): print row
        [Header('SchoolYear'), Header('ID'), Header('Title'), Header('Start'), Header('End')]
        [Text(u'2005'), Text(u'spring'), Text('Spring'), Date(datetime.date(2005, 1, 1)), Date(datetime.date(2005, 1, 30))]
        []
        [Header('Holidays')]
        [Date(datetime.date(2005, 1, 10)), Date(datetime.date(2005, 1, 14))]
        [Date(datetime.date(2005, 1, 17)), Date(datetime.date(2005, 1, 17))]
        []
        [Header('Weekends')]
        [Text('Monday'), Text('Tuesday'), Text('Wednesday'), Text('Thursday'), Text('Friday'), Text('Saturday'), Text('Sunday')]
        [Text(''), Text(''), Text(''), Text(''), Text(''), Text('X'), Text('X')]
        []
        [Header('Working weekends')]
        [Date(datetime.date(2005, 1, 8))]
        [Date(datetime.date(2005, 1, 9))]
    """


def doctest_format_courses():
    """
        >>> app = ISchoolToolApplication(None)
        >>> setUpSchool(app)
        >>> exporter = MegaExporter(app, None)
        >>> for row in exporter.format_courses(): print row
        [Header('School Year'), Header('ID'), Header('Title'), Header('Description'), Header('Local ID'), Header('Government ID'), Header('Credits')]
        [Text(u'2005'), Text(u'c1'), Text('History'), Text(None), Text(None), Text(None), Text(None)]
    """


def doctest_format_persons():
    """
        >>> app = ISchoolToolApplication(None)
        >>> setUpSchool(app)
        >>> exporter = MegaExporter(app, None)
        >>> from pprint import pprint
        >>> for row in exporter.format_persons(): pprint(row)
        [Header('User Name'),
         Header('Prefix'),
         Header('First Name'),
         Header('Middle Name'),
         Header('Last Name'),
         Header('Suffix'),
         Header('Preferred Name'),
         Header('Birth Date'),
         Header('Gender'),
         Header('Password'),
         Header(u'ID'),
         Header(u'Ethnicity'),
         Header(u'Language'),
         Header(u'Place of birth'),
         Header(u'Citizenship')]
        [Text(u'john'),
         Text(None),
         Text('John'),
         Text(None),
         Text('Peterson'),
         Text(None),
         Text(None),
         Date(None),
         Text(None),
         Text(None),
         Text('112323'),
         Text(u'Asian'),
         Text('English'),
         Text('Humptington'),
         Text('US')]
        [Text(u'manager'),
         Text(None),
         Text('Default'),
         Text(None),
         Text('Manager'),
         Text(None),
         Text(None),
         Date(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None)]
        [Text(u'pete'),
         Text(None),
         Text('Pete'),
         Text(None),
         Text('Johnson'),
         Text(None),
         Text(None),
         Date(None),
         Text(None),
         Text(None),
         Text('333655'),
         Text(u'White'),
         Text('Brittish'),
         Text('Providence'),
         Text('UK')]
        [Text(u'teacher'),
         Text(None),
         Text('Mister'),
         Text(None),
         Text('T'),
         Text(None),
         Text(None),
         Date(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None)]
    """


def doctest_format_contact_persons():
    """
        >>> app = ISchoolToolApplication(None)
        >>> setUpSchool(app)
        >>> exporter = MegaExporter(app, None)
        >>> from pprint import pprint
        >>> for row in exporter.format_contact_persons(): pprint(row)
        [Header('ID'),
         Header('Prefix'),
         Header('First Name'),
         Header('Middle Name'),
         Header('Last Name'),
         Header('Suffix'),
         Header('Address line 1'),
         Header('Address line 2'),
         Header('City'),
         Header('State'),
         Header('Country'),
         Header('Postal code'),
         Header('Home phone'),
         Header('Work phone'),
         Header('Mobile phone'),
         Header('Email'),
         Header('Language')]
        [Text('john'),
         Text(''),
         Text(''),
         Text(''),
         Text(''),
         Text(''),
         Text(None),
         Text(None),
         Date(None),
         Date(None),
         Date(None),
         Date(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None)]
        [Text('manager'),
         Text(''),
         Text(''),
         Text(''),
         Text(''),
         Text(''),
         Text(None),
         Text(None),
         Date(None),
         Date(None),
         Date(None),
         Date(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None)]
        [Text('pete'),
         Text(''),
         Text(''),
         Text(''),
         Text(''),
         Text(''),
         Text(None),
         Text(None),
         Date(None),
         Date(None),
         Date(None),
         Date(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None)]
        [Text(u'pete_parent'),
         Text('Ms.'),
         Text('Susan'),
         Text('T.'),
         Text('Johnson'),
         Text('Jr.'),
         Text('1 First St.'),
         Text('Apt. 1'),
         Date('NY'),
         Date('NY'),
         Date('USA'),
         Date('00000'),
         Text('000-0000'),
         Text('111-1111'),
         Text('222-2222'),
         Text('davejohnson@gmail.com'),
         Text('English')]
        [Text('teacher'),
         Text(''),
         Text(''),
         Text(''),
         Text(''),
         Text(''),
         Text(None),
         Text(None),
         Date(None),
         Date(None),
         Date(None),
         Date(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None),
         Text(None)]
    """


def doctest_format_contact_relationships():
    """
        >>> app = ISchoolToolApplication(None)
        >>> setUpSchool(app)
        >>> exporter = MegaExporter(app, None)
        >>> from pprint import pprint
        >>> for row in exporter.format_contact_relationships(): pprint(row)
        [Header('Person ID'), Header('Contact ID'), Header('Relationship')]
        [Text('pete'), Text(u'pete_parent'), Text('parent')]
        [Text('pete'), Text('teacher'), Text('parent')]
    """


class Cell(object):
    def __init__(self, data, style):
        self.data, self.style = data, style


class WorkSheetStub(object):

    def __init__(self):
        self.table = {}

    def write(self, row, col, data, style):
        self.table[row, col] = Cell(data, style)

    def write_merge(self, row1, row2, col1, col2, data, style):
        self.table[row1, col1] = Cell(data, style)

    def format(self):
        dim1 = max([key[0] for key in self.table.keys()]) + 1
        dim2 = max([key[1] for key in self.table.keys()]) + 1

        table = []
        for x in range(dim1):
            l = []
            for y in range(dim2):
                l.append(self.table.get((x, y), Cell("", None)).data)
            table.append(l)
        return format_table(table)


def doctest_MegaExporter_holidays():
    """

        >>> app = ISchoolToolApplication(None)
        >>> sy = ISchoolYearContainer(app)['2005'] = SchoolYear('2005',
        ...                                                     date(2005, 1, 1),
        ...                                                     date(2006, 1, 1))

        >>> term = sy['spring'] = Term("Spring", date(2005, 1, 1), date(2006, 1, 1))
        >>> term.addWeekdays(0, 1, 2, 3, 4, 5, 6)
        >>> exporter = MegaExporter(app, None)
        >>> exporter.calculate_holidays_and_weekdays()
        [[], [], []]

        >>> term.removeWeekdays(0, 1, 2, 3, 4, 5, 6)
        >>> exporter.calculate_holidays_and_weekdays()
        [[], [0, 1, 2, 3, 4, 5, 6], []]

        >>> term.addWeekdays(0, 1, 2, 3, 4)
        >>> exporter.calculate_holidays_and_weekdays()
        [[], [5, 6], []]

        >>> term.add(date(2005, 1, 1))
        >>> term.add(date(2005, 1, 2))
        >>> exporter.calculate_holidays_and_weekdays()
        [[], [5, 6], [datetime.date(2005, 1, 1), datetime.date(2005, 1, 2)]]

        >>> for day in DateRange(date(2005, 12, 18), date(2006, 1, 1)):
        ...     if term.isSchoolday(day): term.remove(day)
        >>> exporter.calculate_holidays_and_weekdays()
        [[(datetime.date(2005, 12, 19), datetime.date(2005, 12, 23)),
          (datetime.date(2005, 12, 26), datetime.date(2005, 12, 30))],
         [5, 6],
         [datetime.date(2005, 1, 1), datetime.date(2005, 1, 2)]]

    """


def doctest_MegaExporter_merge_ranges():
    """

        >>> merge_date_ranges([])
        []

        >>> merge_date_ranges([date(2005, 12, 19)])
        [(datetime.date(2005, 12, 19), datetime.date(2005, 12, 19))]

        >>> merge_date_ranges([date(2005, 12, 19), date(2005, 12, 20)])
        [(datetime.date(2005, 12, 19), datetime.date(2005, 12, 20))]

        >>> merge_date_ranges([date(2005, 12, 19), date(2005, 12, 21)])
        [(datetime.date(2005, 12, 19), datetime.date(2005, 12, 19)),
         (datetime.date(2005, 12, 21), datetime.date(2005, 12, 21))]

        >>> dates = [date(2005, 12, 19),
        ...          date(2005, 12, 20),
        ...          date(2005, 12, 21),
        ...          date(2005, 12, 22),
        ...          date(2005, 12, 23),
        ...          date(2005, 12, 26),
        ...          date(2005, 12, 27),
        ...          date(2005, 12, 28),
        ...          date(2005, 12, 29),
        ...          date(2005, 12, 30)]

        >>> merge_date_ranges(dates)
        [(datetime.date(2005, 12, 19), datetime.date(2005, 12, 23)),
         (datetime.date(2005, 12, 26), datetime.date(2005, 12, 30))]

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_NDIFF)
    suite = doctest.DocTestSuite(optionflags=optionflags,
                                 extraglobs={'provideAdapter': provideStubAdapter,
                                             'provideUtility': provideStubUtility},
                                 setUp=setUp, tearDown=tearDown)
    suite.layer = export_functional_layer
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.app.relationships
"""

import unittest
import doctest

from zope.interface import implements
from zope.component import provideAdapter


def doctest_Instruction():
    r"""Tests for Instruction URIs and methods

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> from schooltool.app.relationships import enforceInstructionConstraints
        >>> setUp()
        >>> import zope.event
        >>> old_subscribers = zope.event.subscribers[:]
        >>> zope.event.subscribers.append(enforceInstructionConstraints)

    We will need some sample persons and sections for the demonstration

        >>> from schooltool.person.person import Person
        >>> from schooltool.course.section import Section
        >>> jonas = Person()
        >>> petras = Person()
        >>> developers = Section()
        >>> admins = Section()

    There are some constraints: Only objects providing ISection can be
    Sections.

        >>> from schooltool.app.relationships import Instruction
        >>> Instruction(instructor=jonas, section=petras)
        Traceback (most recent call last):
        ...
        InvalidRelationship: Sections must provide ISection.

        >>> zope.event.subscribers[:] = old_subscribers
        >>> tearDown()

    """


def doctest_CourseSections():
    r"""Tests for CourseSections relationship

    Lets import the pieces of CourseSections

        >>> from schooltool.app.relationships import CourseSections
        >>> from schooltool.app.relationships import enforceCourseSectionConstraint

    Relationship tests require some setup:

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()
        >>> import zope.event
        >>> old_subscribers = zope.event.subscribers[:]
        >>> zope.event.subscribers.append(enforceCourseSectionConstraint)

    We will need a course and several sections:

        >>> from schooltool.course.course import Course
        >>> from schooltool.course.section import Section
        >>> from schooltool.person.person import Person
        >>> history = Course()
        >>> section1 = Section(title = "section1")
        >>> section2 = Section(title = "section2")
        >>> section3 = Section(title = "section3")
        >>> section4 = Section(title = "section4")
        >>> person = Person()

    Our course doesn't have any sections yet:

        >>> for section in history.sections:
        ...     print section

    Lets add one:

        >>> history.sections.add(section1)
        >>> for section in history.sections:
        ...     print section.title
        section1

    Lets try to add a person to the course:

        >>> history.sections.add(person)
        Traceback (most recent call last):
        ...
        InvalidRelationship: Sections must provide ISection.

    Lets try to add a course to the course:

        >>> algebra = Course()
        >>> history.sections.add(algebra)
        Traceback (most recent call last):
        ...
        InvalidRelationship: Sections must provide ISection.

    No luck, you can only add sections:

        >>> history.sections.add(section2)
        >>> history.sections.add(section3)
        >>> for section in history.sections:
        ...     print section.title
        section1
        section2
        section3

    You can use the Relationship to relate sections and courses:

        >>> CourseSections(course=history, section=section4)
        >>> for section in history.sections:
        ...     print section.title
        section1
        section2
        section3
        section4

    That's it:

        >>> zope.event.subscribers[:] = old_subscribers
        >>> tearDown()

    """


def getSectionCalendar(section):
    return section.calendar


from zope.annotation.interfaces import IAttributeAnnotatable
class CalendarStub(object):
    implements(IAttributeAnnotatable)
    def __init__(self, section):
        self.title = section.title


def doctest_updateInstructorCalendars():
    r"""
        >>> from schooltool.app.relationships import updateInstructorCalendars
        >>> from schooltool.app.relationships import URIInstruction
        >>> from schooltool.app.relationships import URISection, URIInstructor
        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> from schooltool.course.interfaces import ISection
        >>> from schooltool.relationship.interfaces import \
        ...                                         IRelationshipAddedEvent
        >>> from schooltool.relationship.interfaces import \
        ...                                         IRelationshipRemovedEvent
        >>> from schooltool.person.person import Person
        >>> from schooltool.course.section import Section
        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

        >>> provideAdapter(getSectionCalendar,
        ...                adapts=(ISection,),
        ...                provides=ISchoolToolCalendar)

        >>> from schooltool.app.overlay import ICalendarOverlayInfo

        >>> class AddEventStub(dict):
        ...     rel_type = URIInstruction
        ...     implements(IRelationshipAddedEvent)

        >>> class RemoveEventStub(dict):
        ...     rel_type = URIInstruction
        ...     implements(IRelationshipRemovedEvent)

        >>> class OtherEventStub(dict):
        ...     rel_type = URIInstruction

        >>> person = Person()
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []
        >>> section = Section(title="SectionA")
        >>> section.calendar = CalendarStub(section)

    When the person is made the instructor of a section the sections calendar
    is added to the overlaid calendars:

        >>> add = AddEventStub()
        >>> add[URIInstructor] = person
        >>> add[URISection] = section
        >>> updateInstructorCalendars(add)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        ['SectionA']

    The calendar of the section is visible.

        >>> [cal.show for cal in person.overlaid_calendars]
        [True]

    The calendar is removed when the instructor is no longer in the section:

        >>> remove = RemoveEventStub()
        >>> remove[URIInstructor] = person
        >>> remove[URISection] = section
        >>> updateInstructorCalendars(remove)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []

    If a person allready has that calendar nothing changes:

        >>> sectionb = Section(title="SectionB")
        >>> sectionb.calendar = CalendarStub(sectionb)
        >>> person.overlaid_calendars.add(sectionb.calendar)
        <...CalendarOverlayInfo object at ...>
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        ['SectionB']

        >>> add = AddEventStub()
        >>> add[URIInstructor] = person
        >>> add[URISection] = sectionb
        >>> updateInstructorCalendars(add)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        ['SectionB']

    If the person removes the calendar manually, that's ok:

        >>> person.overlaid_calendars.remove(sectionb.calendar)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []


        >>> remove = RemoveEventStub()
        >>> remove[URIInstructor] = person
        >>> remove[URISection] = sectionb
        >>> updateInstructorCalendars(remove)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []

    Events that aren't RelationshipAdded/Removed are ignored:

        >>> other = OtherEventStub()
        >>> other[URIInstructor] = person
        >>> other[URISection] = sectionb
        >>> updateInstructorCalendars(other)

        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []


        >>> tearDown()

    """


def doctest_updateStudentCalendars():
    r"""
        >>> from schooltool.app.relationships import updateStudentCalendars
        >>> from schooltool.app.membership import URIMembership
        >>> from schooltool.app.membership import URIGroup, URIMember
        >>> from schooltool.relationship.interfaces import \
        ...                                         IRelationshipAddedEvent
        >>> from schooltool.relationship.interfaces import \
        ...                                         IRelationshipRemovedEvent
        >>> from schooltool.course.section import Section
        >>> from schooltool.person.person import Person
        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

        >>> from schooltool.course.interfaces import ISection
        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> provideAdapter(getSectionCalendar,
        ...                adapts=(ISection,),
        ...                provides=ISchoolToolCalendar)

        >>> from schooltool.app.overlay import ICalendarOverlayInfo

        >>> class AddEventStub(dict):
        ...     rel_type = URIMembership
        ...     implements(IRelationshipAddedEvent)

        >>> class RemoveEventStub(dict):
        ...     rel_type = URIMembership
        ...     implements(IRelationshipRemovedEvent)

        >>> class OtherEventStub(dict):
        ...     rel_type = URIMembership

        >>> person = Person('p')
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []
        >>> section = Section(title="SectionA")
        >>> section.calendar = CalendarStub(section)

    When the person is made a member of a section the sections calendar
    is added to the overlaid calendars:

        >>> add = AddEventStub()
        >>> add[URIMember] = person
        >>> add[URIGroup] = section
        >>> updateStudentCalendars(add)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        ['SectionA']

    The calendar of the section is visible.

        >>> [cal.show for cal in person.overlaid_calendars]
        [True]

    The calendar is removed when the person is no longer in the section:

        >>> remove = RemoveEventStub()
        >>> remove[URIMember] = person
        >>> remove[URIGroup] = section
        >>> updateStudentCalendars(remove)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []

    If a person already has that calendar nothing changes:

        >>> sectionb = Section(title="SectionB")
        >>> sectionb.calendar = CalendarStub(sectionb)
        >>> person.overlaid_calendars.add(sectionb.calendar)
        <...CalendarOverlayInfo object at ...>
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        ['SectionB']

        >>> add = AddEventStub()
        >>> add[URIMember] = person
        >>> add[URIGroup] = sectionb
        >>> updateStudentCalendars(add)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        ['SectionB']

    If the person removes the calendar manually, that's ok for now, we may
    want to take away this ability later.

        >>> person.overlaid_calendars.remove(sectionb.calendar)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []

        >>> remove = RemoveEventStub()
        >>> remove[URIMember] = person
        >>> remove[URIGroup] = sectionb
        >>> updateStudentCalendars(remove)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []

    Events that aren't RelationshipAdded/Removed are ignored:

        >>> other = OtherEventStub()
        >>> other[URIMember] = person
        >>> other[URIGroup] = sectionb
        >>> updateStudentCalendars(other)

        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []

    If you add a person to a Group that isn't a section, nothing happens, the
    user will have to overlay the calendar manually:

        >>> from schooltool.group.group import Group
        >>> person = Person('p2')
        >>> group = Group()
        >>> add = AddEventStub()
        >>> add[URIMember] = person
        >>> add[URIGroup] = group
        >>> updateStudentCalendars(add)
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []

    You can add a group to a section and it's members overlay list will be
    updated:

        >>> student = Person('p3')
        >>> [cal.calendar.title for cal in person.overlaid_calendars]
        []
        >>> freshmen = Group()
        >>> freshmen.members.add(student)
        >>> section = Section("Freshmen Math")
        >>> section.calendar = CalendarStub(section)
        >>> add = AddEventStub()
        >>> add[URIMember] = freshmen
        >>> add[URIGroup] = section
        >>> updateStudentCalendars(add)
        >>> [cal.calendar.title for cal in student.overlaid_calendars]
        ['Freshmen Math']

        >>> [cal.show for cal in student.overlaid_calendars]
        [True]

        >>> remove = RemoveEventStub()
        >>> remove[URIMember] = freshmen
        >>> remove[URIGroup] = section
        >>> updateStudentCalendars(remove)
        >>> [cal.calendar.title for cal in student.overlaid_calendars]
        []

        >>> tearDown()

    """


def test_suite():
    return unittest.TestSuite([
                doctest.DocTestSuite(optionflags=doctest.ELLIPSIS),
                doctest.DocTestSuite('schooltool.app.relationships',
                                     optionflags=doctest.ELLIPSIS),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool skin.
"""

from zope.interface import Interface
from zope.interface import implements
from zope.schema import Object
from zope.viewlet.interfaces import IViewletManager
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.app.browser.interfaces import IEventForDisplay
from schooltool.skin.skin import OrderedViewletManager
from schooltool.resource.interfaces import IBookingCalendarEvent


class ICalendarEventViewletManager(IViewletManager):
    """Provides a viewlet hook for daily calendar events."""


class ICalendarEventContext(Interface):
    """Schema for attributes required by CalendarEventViewletManager."""

    event = Object(IEventForDisplay)


class CalendarEventBookingViewlet(object):
    """
    This is the view class for the booking viewlet on the CalendarEventView
    """

    def bookingUrl(self, resource_id):
        url = absoluteURL(self.context, self.request)
        url = "%s/book_one_resource.html?resource_id=%s" % (url, resource_id)
        event = self.manager.event.context
        if IBookingCalendarEvent.providedBy(event):
            url = "%s&event_id=%s" % (url, event.unique_id)
        else:
            url = "%s&start_date=%s&start_time=%s&duration=%s&title=%s" % (
                url, event.dtstart.date(), event.dtstart.time(),
                event.duration.seconds, event.title)

        return url

    def listResources(self):
        return [{'id': resource.__name__, 'title': resource.title}
                for resource in self.manager.event.context.resources]


class CalendarEventViewletManager(OrderedViewletManager):
    """Viewlet manager for displaying of additional event information."""

    implements(ICalendarEventContext)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Resource booking for sections.
"""

from schooltool.relationship.uri import URIObject
from schooltool.relationship.relationship import RelationshipSchema


URISectionBooking = URIObject('http://schooltool.org/ns/sectionbooking',
                              'Section booking',
                              'The section booking relationship.')
URISection = URIObject('http://schooltool.org/ns/sectionbooking/section',
                       'Section', 'A role of a section.')
URIResource = URIObject('http://schooltool.org/ns/sectionbooking/resource',
                        'Resource', 'The role of a booked resource.')


SectionBooking = RelationshipSchema(URISectionBooking,
                                    section=URISection,
                                    resource=URIResource)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for schooltool error views.
"""

import unittest
import doctest

from zope.publisher.browser import TestRequest
from zope.browser.interfaces import ISystemErrorView


def doctest_ErrorView():
    r"""Test for ErrorView

        >>> from schooltool.skin.error import ErrorView
        >>> exception = RuntimeError("some kind of error")
        >>> request = TestRequest()
        >>> view = ErrorView(exception, request)

    The view is a system error view, thus causing the zope publication to
    log exceptions via the SiteError logger.

        >>> ISystemErrorView.providedBy(view)
        True
        >>> view.isSystemError()
        True

    view.index is a page template, normally specified via ZCML

        >>> view.index = lambda: 'rendered page'

    Rendering the view sets the status code to 500.

        >>> print view()
        rendered page

        >>> request.response.getStatus()
        500

    view.traceback uses sys.exc_info, and apparently we have to test it inside
    an except clause

        >>> try:
        ...     raise exception
        ... except:
        ...     print view.traceback
        File "<span class="filename">&lt;doctest...doctest_ErrorView[...]&gt;</span>",
          line <span class="lineno">2</span>, in <span class="method">...</span>
          <span class="source">raise exception</span>

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for customisation of SchoolTool security policy.
"""
import unittest
import doctest

from zope.interface import implements
from zope.publisher.browser import TestRequest


def doctest_AccessControlView():
    r"""Tests for AccessControlView.

    The view should display a list of all settings provided by the
    AccessControlCustomisations of the SchoolToolApplication:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.securitypolicy.interfaces import IAccessControlCustomisations
        >>> class CustomisationsStub(object):
        ...     implements(IAccessControlCustomisations)
        ...     def __iter__(self):
        ...         return iter(['Access', 'control', 'settings'])

        >>> class AppStub(object):
        ...     implements(ISchoolToolApplication)
        ...     def __conform__(self, iface):
        ...         if iface == IAccessControlCustomisations:
        ...             return CustomisationsStub()

        >>> from zope.component import provideAdapter
        >>> provideAdapter(lambda context: AppStub(),
        ...                adapts=[None],
        ...                provides=ISchoolToolApplication)

        >>> from schooltool.securitypolicy.browser.views import AccessControlView
        >>> view = AccessControlView(None, None)
        >>> view.settings()
        ['Access', 'control', 'settings']

    """


def doctest_AccessControlView_update():
    """Tests for AccessControlView.update()

        >>> from schooltool.securitypolicy.browser.views import AccessControlView
        >>> request = TestRequest()
        >>> view = AccessControlView(None, request)

    If there is nothing in the form, update does nothing:

        >>> view.update()

    If we click on Update, it should update all the settings with data
    in the form:

        >>> class SettingStub(object):
        ...     def __init__(self, key):
        ...         self.key = key
        ...     def setValue(self, value):
        ...         print "Setting value of '%s' to '%s'" % (self.key, value)
        >>> view.settings = lambda: [SettingStub('one'), SettingStub('two')]
        >>> request.form = {'UPDATE_SUBMIT': '', 'setting.one': 'checked'}
        >>> view.update()
        Setting value of 'one' to 'True'
        Setting value of 'two' to 'False'

    """


def test_suite():
    return unittest.TestSuite([
            doctest.DocTestSuite(optionflags=doctest.ELLIPSIS |
                                             doctest.NORMALIZE_WHITESPACE)])




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import

import bottle
import types
try:
    import json
except ImportError:
    import simplejson as json

import zope.i18nmessageid.message
import zope.configuration.config
import zope.configuration.xmlconfig
from zope.i18n import translate
from zope.interface import directlyProvides
from zope.publisher.http import IHTTPRequest

from schooltool.app.main import SchoolToolMachinery, setLanguage
from schooltool.task.state import TaskReadState


not_translatable = object()


def iter_translate(o, markers=None):
    for item in o:
        text = inplace_translate(item, markers=markers)
        yield item if text is not_translatable else text


def inplace_translate(o, markers=None):
    oid = id(o)
    if markers is not None:
        if oid in markers:
            return markers[oid]
        else:
            markers[oid] = not_translatable
    else:
        markers = {oid: not_translatable}

    if isinstance(o, zope.i18nmessageid.message.Message):
        try:
            markers[oid] = translate(o, context=bottle.request)
        except TypeError:
            markers[oid] = o
    elif isinstance(o, (list, tuple)):
        if markers is None:
            markers = {}
        markers[oid] = o.__class__(iter_translate(o, markers=markers))
    elif isinstance(o, dict):
        for k in o:
            t = inplace_translate(o[k])
            if t is not_translatable:
                continue
            o[k] = t
        markers[oid] = o
    return markers[oid]


result_app = bottle.Bottle()


class JSONEncoder(json.JSONEncoder):

    def default(self, ob):
        if type(ob) == types.GeneratorType:
            return list(ob)
        if isinstance(ob, Exception):
            return repr(ob)
        return json.JSONEncoder.default(self, ob)


def encode_json(obj):
    encoder = JSONEncoder(sort_keys=True, separators=(',', ':'))
    return encoder.encode(obj)


def status_dict(status):
    result = {}
    for attr in ('pending', 'in_progress', 'committing',
                 'finished', 'failed', 'succeeded'):
        result[attr] = getattr(status, attr)
    return result


def make_task_result(task_id):
    status = TaskReadState(task_id)
    result = {
        'internal_state': status.state,
        'status': status_dict(status),
        'info': status.info,
        'traceback': status.traceback,
        }
    return result


@result_app.route('/<task_id>')
def fetch_full(task_id=None):
    if task_id is None:
        raise bottle.HTTPError(404, "Not found: %r" % bottle.request.url)
    result = make_task_result(task_id)
    directlyProvides(bottle.request, IHTTPRequest)
    translated_result = inplace_translate(result)
    encoded_result = encode_json(translated_result)
    bottle.response.set_header('Content-Type', 'application/json')
    return encoded_result


class ResultServerMachinery(SchoolToolMachinery):

    def configureComponents(self):
        context = zope.configuration.config.ConfigurationMachine()
        if self.config.devmode:
            context.provideFeature('devmode')
        zope.configuration.xmlconfig.registerCommonDirectives(context)
        context = zope.configuration.xmlconfig.file(
            self.config.result_server_definition, context=context)

    def configure(self, config_file):
        self.config, self.handler = self.readConfig(config_file)
        self.configureComponents()
        setLanguage(self.config.lang)

    @property
    def app(self):
        global result_app
        return result_app



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish skin common helpers.
"""
import urllib

import zope.security
import zope.container.interfaces


class Empty(object):

    __name__ = None
    __parent__ = None

    def __init__(self, context, request, *args, **kw):
        self.__parent__ = context
        self.context = context
        self.request = request

    def update(self):
        pass

    def render(self, *args, **kw):
        return ''

    def __call__(self, *args, **kw):
        return ''


class EmptyContent(Empty):

    def __init__(self, context, request, view, **kw):
        Empty.__init__(self, context, request, **kw)
        self.__parent__ = view
        self.view = view


class EmptyViewlet(EmptyContent):

    def __init__(self, context, request, view, manager, **kw):
        EmptyContent.__init__(self, context, request, view, **kw)
        self.__parent__ = manager
        self.manager = manager


def canView(context):
    return zope.security.checkPermission('schooltool.view', context)


def canEdit(context):
    return zope.security.checkPermission('schooltool.edit', context)


def canDelete(context):
    container = context.__parent__
    if not zope.container.interfaces.IWriteContainer.providedBy(container):
        raise NotImplementedError()
    # XXX: if context has dependents, deletion may be prevented
    return zope.security.canAccess(container, '__delitem__')


def hasPermission(context, permission):
    return zope.security.checkPermission(permission, context)


def quoteFilename(filename):
    if not filename:
        return filename
    if type(filename) is unicode:
        encoded = filename.encode('UTF-8')
    else:
        encoded = str(filename)
    return urllib.quote(encoded)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for basic person security.
"""
import unittest
import doctest

from schooltool.relationship.tests import setUp, tearDown


def doctest_PersonAdvisorsCrowd():
    """Tests for PersonAdvisorsCrowd.

    We'll create an advisor and two students.

        >>> from schooltool.basicperson.person import BasicPerson
        >>> student1 = BasicPerson("student", "Student", "One")
        >>> student2 = BasicPerson("student", "Student", "Two")
        >>> advisor = BasicPerson("advisor", "Advisor", "One")

    The advisor will only advise the first student.

        >>> student1.advisors.add(advisor)

    Hence, the advisor will belong to the first student's advisors crowd and
    not the second.
    
        >>> from schooltool.basicperson.security import PersonAdvisorsCrowd
        >>> PersonAdvisorsCrowd(student1).contains(advisor)
        True
        >>> PersonAdvisorsCrowd(student2).contains(advisor)
        False

    """


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    return doctest.DocTestSuite(optionflags=optionflags,
                                setUp=setUp, tearDown=tearDown)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for calendaring related timetabling views.
"""
import unittest
from pytz import timezone
from datetime import timedelta, time, date, datetime

from zope.publisher.browser import TestRequest

from schooltool.common import parse_datetime
from schooltool.person.interfaces import IPersonPreferences
from schooltool.person.person import Person
from schooltool.term.interfaces import ITermContainer
from schooltool.testing.util import NiceDiffsMixin
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.tests.test_cal import dt
from schooltool.app.browser.testing import layeredTestTearDown
from schooltool.app.browser.testing import layeredTestSetup
from schooltool.app.browser.testing import makeLayeredSuite
from schooltool.app.testing import app_functional_layer

try:
    from schooltool.timetable.interfaces import ITimetableSchemaContainer
    from schooltool.timetable.schema import TimetableSchema
    from schooltool.timetable.model import SequentialDaysTimetableModel
    from schooltool.timetable import SchooldaySlot
    from schooltool.timetable import SchooldayTemplate
except:
    pass # XXX: tests not refactored yet

class TestDailyTimetableCalendarRowsView(NiceDiffsMixin, unittest.TestCase):

    def setUp(self):
        layeredTestSetup()
        app = ISchoolToolApplication(None)
        self.person = app['persons']['person'] = Person('person')

        # set up schoolyear
        from schooltool.schoolyear.schoolyear import SchoolYear
        from schooltool.schoolyear.interfaces import ISchoolYearContainer
        ISchoolYearContainer(app)['2004'] = SchoolYear("2004", date(2004, 9, 1), date(2004, 12, 31))

        # set up the timetable schema
        days = ['A', 'B', 'C']
        schema = self.createSchema(days,
                                   ['1', '2', '3', '4'],
                                   ['1', '2', '3', '4'],
                                   ['1', '2', '3', '4'])
        schema.timezone = 'Europe/London'
        template = SchooldayTemplate()
        template.add(SchooldaySlot(time(8, 0), timedelta(hours=1)))
        template.add(SchooldaySlot(time(10, 15), timedelta(hours=1)))
        template.add(SchooldaySlot(time(11, 30), timedelta(hours=1)))
        template.add(SchooldaySlot(time(12, 30), timedelta(hours=2)))
        schema.model = SequentialDaysTimetableModel(days, {None: template})

        ITimetableSchemaContainer(app)['default'] = schema

        # set up terms
        from schooltool.term.term import Term
        terms = ITermContainer(app)
        terms['term'] = term = Term("Some term", date(2004, 9, 1),
                                    date(2004, 12, 31))
        term.add(date(2004, 11, 5))

    def tearDown(self):
        layeredTestTearDown()

    def createSchema(self, days, *periods_for_each_day):
        """Create a timetable schema."""
        from schooltool.timetable.schema import TimetableSchemaDay
        schema = TimetableSchema(days, title="A Schema")
        for day, periods in zip(days, periods_for_each_day):
            schema[day] = TimetableSchemaDay(list(periods))
        return schema

    def test_calendarRows(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView
        from schooltool.app.security import Principal

        request = TestRequest()
        principal = Principal('person', 'Some person', person=self.person)
        request.setPrincipal(principal)
        view = DailyTimetableCalendarRowsView(ISchoolToolCalendar(self.person), request)
        result = list(view.calendarRows(date(2004, 11, 5), 8, 19, events=[]))

        expected = [("1", dt('08:00'), timedelta(hours=1)),
                    ("9:00", dt('09:00'), timedelta(hours=1)),
                    ("10:00", dt('10:00'), timedelta(minutes=15)),
                    ("2", dt('10:15'), timedelta(hours=1)),
                    ("11:15", dt('11:15'), timedelta(minutes=15)),
                    ("3", dt('11:30'), timedelta(hours=1)),
                    ("4", dt('12:30'), timedelta(hours=2)),
                    ("14:30", dt('14:30'), timedelta(minutes=30)),
                    ("15:00", dt('15:00'), timedelta(hours=1)),
                    ("16:00", dt('16:00'), timedelta(hours=1)),
                    ("17:00", dt('17:00'), timedelta(hours=1)),
                    ("18:00", dt('18:00'), timedelta(hours=1))]

        self.assertEquals(result, expected)

    def test_calendarRows_otherTZ(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView
        from schooltool.app.security import Principal

        request = TestRequest()
        principal = Principal('person', 'Some person', person=self.person)
        request.setPrincipal(principal)
        view = DailyTimetableCalendarRowsView(ISchoolToolCalendar(self.person), request)

        km = timezone('Asia/Kamchatka')
        view.getPersonTimezone = lambda: km

        result = list(view.calendarRows(date(2004, 11, 5), 8, 19, events=[]))

        kmdt = lambda arg: km.localize(parse_datetime('2004-11-05 %s:00' %
                                                      arg))

        expected = [('8:00', kmdt('8:00'), timedelta(0, 3600)),
                    ('9:00', kmdt('9:00'), timedelta(0, 3600)),
                    ('10:00', kmdt('10:00'),timedelta(0, 3600)),
                    ('11:00', kmdt('11:00'),timedelta(0, 3600)),
                    ('12:00', kmdt('12:00'),timedelta(0, 3600)),
                    ('13:00', kmdt('13:00'),timedelta(0, 3600)),
                    ('14:00', kmdt('14:00'),timedelta(0, 3600)),
                    ('15:00', kmdt('15:00'),timedelta(0, 3600)),
                    ('16:00', kmdt('16:00'),timedelta(0, 3600)),
                    ('17:00', kmdt('17:00'),timedelta(0, 3600)),
                    ('18:00', kmdt('18:00'),timedelta(0, 3600)),
                    ('19:00', kmdt('19:00'),timedelta(0, 3600)),
                    ('1',  kmdt("20:00"), timedelta(0, 3600)),
                    ('21:00', kmdt("21:00"), timedelta(0, 4500)),
                    ('2',  kmdt("22:15"), timedelta(0, 3600)),
                    ('23:15', kmdt("23:15"), timedelta(0, 900)),
                    ('3', kmdt("23:30"), timedelta(0, 1800))]

        self.assertEquals(result, expected)

    def test_calendarRows_no_periods(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView
        from schooltool.person.preference import getPersonPreferences
        from schooltool.app.security import Principal

        prefs = getPersonPreferences(self.person)
        prefs.cal_periods = False # do not show periods
        request = TestRequest()
        principal = Principal('person', 'Some person', person=self.person)
        request.setPrincipal(principal)
        view = DailyTimetableCalendarRowsView(ISchoolToolCalendar(self.person), request)

        result = list(view.calendarRows(date(2004, 11, 5), 8, 19, events=[]))

        expected = [("%d:00" % i, dt('%d:00' % i), timedelta(hours=1))
                    for i in range(8, 19)]
        self.assertEquals(result, expected)

    def test_calendarRows_default(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView
        request = TestRequest()
        # do not set the principal
        view = DailyTimetableCalendarRowsView(ISchoolToolCalendar(self.person), request)
        result = list(view.calendarRows(date(2004, 11, 5), 8, 19, events=[]))

        # the default is not to show periods
        expected = [("%d:00" % i, dt('%d:00' % i), timedelta(hours=1))
                    for i in range(8, 19)]
        self.assertEquals(result, expected)

    def test_getPersonTimezone(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView

        request = TestRequest()
        view = DailyTimetableCalendarRowsView(ISchoolToolCalendar(self.person), request)

        # when there is no principal - the default timezone should be
        # returned
        self.assertEquals(view.getPersonTimezone(), timezone('UTC'))

    def test_getPeriods(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView

        request = TestRequest()
        view = DailyTimetableCalendarRowsView(ISchoolToolCalendar(self.person), request)

        # if no user has logged we should get an empty list
        self.assertEquals(view.getPeriods(date(2005, 1, 1)), [])

        # same if our user doesn't want to see periods in his calendar
        request.setPrincipal(self.person)
        IPersonPreferences(self.person).cal_periods = False
        self.assertEquals(view.getPeriods(date(2005, 1, 1)), [])

        # if currently logged in user wants to see periods, the
        # parameter is passed to getPeriodsForDay method.
        view.getPeriodsForDay = lambda cursor: ("Yep", cursor)
        IPersonPreferences(self.person).cal_periods = True
        self.assertEquals(view.getPeriods(date(2005, 1, 1)),
                          ("Yep", date(2005, 1, 1)))


class TestDailyTimetableCalendarRowsView_getPeriodsForDay(NiceDiffsMixin,
                                                 unittest.TestCase):

    def setUp(self):
        layeredTestSetup()
        app = ISchoolToolApplication(None)

        from schooltool.schoolyear.schoolyear import SchoolYear
        from schooltool.schoolyear.interfaces import ISchoolYearContainer
        ISchoolYearContainer(app)['2004-2005'] = SchoolYear("2004-2005", date(2004, 9, 1), date(2005, 8, 1))

        from schooltool.term.term import Term
        self.term1 = Term('Sample', date(2004, 9, 1), date(2004, 12, 20))
        self.term1.schooldays = [('A', time(9,0), timedelta(minutes=115)),
                                 ('B', time(11,0), timedelta(minutes=115)),
                                 ('C', time(13,0), timedelta(minutes=115)),
                                 ('D', time(15,0), timedelta(minutes=115)),]
        self.term2 = Term('Sample', date(2005, 1, 1), date(2005, 6, 1))
        self.term2.schooldays = []
        terms = ITermContainer(app)
        terms['2004-fall'] = self.term1
        terms['2005-spring'] = self.term2

        class TimetableModelStub:
            def periodsInDay(this, schooldays, ttschema, date):
                if date not in schooldays:
                    raise "This date is not in the current term!"
                if ttschema == self.tt:
                    return schooldays.schooldays
                else:
                    return []

        tt = TimetableSchema([])
        tt.model = TimetableModelStub()
        tt.timezone = 'Europe/London'
        self.tt = tt
        ttschemas = ITimetableSchemaContainer(app)
        ttschemas['default'] = tt
        self.app = app

    def tearDown(self):
        layeredTestTearDown()

    def test_getPeriodsForDay_sameTZ(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView

        view = DailyTimetableCalendarRowsView(None, TestRequest())
        uk = timezone('Europe/London')
        view.getPersonTimezone = lambda: uk
        delta = timedelta(minutes=115)
        ukdt = lambda *args: uk.localize(datetime(*args))
        self.assertEquals(view.getPeriodsForDay(date(2004, 10, 14)),
                          [('A', ukdt(2004, 10, 14,  9, 0), delta),
                           ('B', ukdt(2004, 10, 14, 11, 0), delta),
                           ('C', ukdt(2004, 10, 14, 13, 0), delta),
                           ('D', ukdt(2004, 10, 14, 15, 0), delta)])

        # However, if there is no time period, we return []
        self.assertEquals(view.getPeriodsForDay(date(2005, 10, 14)),
                          [])

        # If there is no timetable schema, we return []
        ITimetableSchemaContainer(self.app).default_id = None
        self.assertEquals(view.getPeriodsForDay(date(2004, 10, 14)),
                          [])

    def test_getPeriodsForDay_otherTZ(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView

        view = DailyTimetableCalendarRowsView(None, TestRequest())
        # Auckland is 12h ahead of London
        view.getPersonTimezone = lambda: timezone('Pacific/Auckland')
        uk = timezone('Europe/London')

        delta = timedelta(minutes=115)
        td = timedelta
        ukdt = lambda *args: uk.localize(datetime(*args))
        self.assertEquals(view.getPeriodsForDay(date(2004, 10, 14)),
                          [('B', ukdt(2004, 10, 13, 12, 0), td(minutes=55)),
                           ('C', ukdt(2004, 10, 13, 13, 0), delta),
                           ('D', ukdt(2004, 10, 13, 15, 0), delta),
                           ('A', ukdt(2004, 10, 14, 9, 0), delta),
                           ('B', ukdt(2004, 10, 14, 11, 0), td(minutes=60))])

        # However, if there is no time period, we return []
        self.assertEquals(view.getPeriodsForDay(date(2005, 10, 14)),
                          [])

        # If there is no timetable schema, we return []
        ITimetableSchemaContainer(self.app).default_id = None
        self.assertEquals(view.getPeriodsForDay(date(2004, 10, 14)),
                          [])

    def test_getPeriodsForLastDayOfTerm(self):
        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView

        view = DailyTimetableCalendarRowsView(None, TestRequest())
        # We need the start date and the end date different
        view.getPersonTimezone = lambda: timezone('America/Chicago')
        self.assertEquals(view.getPeriodsForDay(date(2005, 6, 1)), [])

    def test_getPeriodsForDayBetweenTerms(self):
        from schooltool.term.term import Term
        term3 = Term('Sample', date(2005, 6, 2), date(2005, 8, 1))
        ITermContainer(self.app)['2005-autumn'] = term3
        term3.schooldays = [('A', time(9,0), timedelta(minutes=115))]
        self.term2.schooldays = [('B', time(10,0), timedelta(minutes=115))]

        from schooltool.timetable.browser.cal import DailyTimetableCalendarRowsView

        view = DailyTimetableCalendarRowsView(None, TestRequest())
        # We need the start date and the end date different
        view.getPersonTimezone = lambda: timezone('America/Chicago')
        self.assertEquals(len(view.getPeriodsForDay(date(2005, 6, 1))), 1)


def test_suite():
    suite = unittest.TestSuite()
    # XXX: tests not refactored yet
    #suite.addTest(makeLayeredSuite(TestDailyTimetableCalendarRowsView,
    #                               app_functional_layer))
    #suite.addTest(makeLayeredSuite(TestDailyTimetableCalendarRowsView_getPeriodsForDay,
    #                               app_functional_layer))
    return suite



#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool application views.
"""

import zope.schema
from zope.proxy import sameProxiedObjects
from zope.interface import implements

from schooltool.common import simple_form_key


def vocabulary(choices):
    """Create a SimpleVocabulary from a list of values and titles.

    >>> v = vocabulary([('value1', u"Title for value1"),
    ...                 ('value2', u"Title for value2")])
    >>> for term in v:
    ...   print term.value, '|', term.token, '|', term.title
    value1 | value1 | Title for value1
    value2 | value2 | Title for value2

    """
    return zope.schema.vocabulary.SimpleVocabulary(
        [zope.schema.vocabulary.SimpleTerm(v, title=t) for v, t in choices])


def vocabulary_titled(items):
    """Create a SimpleVocabulary from a list of objects having __name__ and
    title attributes.  Such items are common in many containers in SchoolTool.

    >>> class Item(object):
    ...     def __init__(self, name, title):
    ...         self.__name__ = name
    ...         self.title = title
    ...     def __repr__(self):
    ...         return '<Item "%s">' % self.title

    >>> v = vocabulary_titled([Item(u'thevalue1', u"Title one"),
    ...                        Item(u'\xc5\xa0amas', u"Title two")])
    >>> for term in v:
    ...   print term.value, '|', term.token, '|', term.title
    <Item "Title one"> | thevalue1- | Title one
    <Item "Title two"> | amas-uea5t | Title two

    """
    return zope.schema.vocabulary.SimpleVocabulary(
        [zope.schema.vocabulary.SimpleTerm(
                    item,
                    token=unicode(item.__name__).encode('punycode'),
                    title=item.title)
         for item in items])


class TitledContainerItemVocabulary(object):
    """Vocabulary of titled items in a container."""
    implements(zope.schema.interfaces.IVocabularyTokenized)

    def __init__(self, context):
        self.context = context

    @property
    def container(self):
        raise NotImplementedError("Don't know how to get the container from context")

    def __len__(self):
        return len(self.container)

    def __contains__(self, item):
        return sameProxiedObjects(item, self.container.get(item.__name__))

    def __iter__(self):
        for item in self.container.values():
            yield self.getTerm(item)

    def getTermByToken(self, token):
        terms = [self.getTerm(item) for item in self.container.values()]
        by_token = dict([(term.token, term) for term in terms])
        if token not in by_token:
            raise LookupError(token)
        return by_token[token]

    def getTerm(self, item):
        return zope.schema.vocabulary.SimpleTerm(
            item,
            token=simple_form_key(item),
            title=self.getTitle(item))

    def getTitle(self, item):
        return item.title
